from unittest import mock

from cryosparc.api import APIClient
from cryosparc.controllers.job import Job, JobController
from cryosparc.controllers.project import Project
from cryosparc.models.job_spec import Params
from cryosparc.tools import CryoSPARC


def test_create_job_basic(cs: CryoSPARC, project: Project, mock_new_job: Job):
    assert isinstance(mock_create_endpoint := APIClient.jobs.create, mock.Mock)
    mock_create_endpoint.return_value = mock_new_job

    job = cs.create_job(project.uid, "W1", "homo_abinit")
    assert isinstance(job, JobController)
    assert job.uid == mock_new_job.uid
    assert len(job.model.spec.params.model_dump(exclude_defaults=True, exclude_none=True)) == 0
    mock_create_endpoint.assert_called_once_with(
        project.uid, "W1", type="homo_abinit", title="", description="", params={}
    )


def test_create_job_connect_params(
    cs: CryoSPARC,
    project: Project,
    mock_params: Params,
    mock_new_job_with_params: Job,
    mock_new_job_with_connection: Job,
):
    assert isinstance(mock_create_endpoint := APIClient.jobs.create, mock.Mock)
    assert isinstance(mock_connect_endpoint := APIClient.jobs.connect, mock.Mock)
    mock_create_endpoint.return_value = mock_new_job_with_params
    mock_connect_endpoint.return_value = mock_new_job_with_connection
    job = cs.create_job(
        project.uid,
        "W1",
        "homo_abinit",
        connections={"particles": ("J41", "particles")},
        params=mock_params.model_dump(),
    )
    assert isinstance(job, JobController)
    assert job.uid == mock_new_job_with_connection.uid
    assert job.model.spec.params == mock_params
    assert len(job.model.spec.inputs.root["particles"].connections) == 1
    mock_create_endpoint.assert_called_once_with(
        project.uid, "W1", type="homo_abinit", title="", description="", params=mock_params.model_dump()
    )
    mock_connect_endpoint.assert_called_once_with(
        project.uid, job.uid, "particles", source_job_uid="J41", source_output_name="particles"
    )



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import List

from pydantic import BaseModel


class ResourceSlots(BaseModel):
    """
    Available compute resources on a worker node that are allocated to a job
    when it runs.
    """

    CPU: List[int] = []
    """
    List of available CPU core numbers. For example, if worker has 4 CPU
    cores allocated, is set to ``[0, 1, 2, 3]``.
    """
    GPU: List[int] = []
    """
    List of available GPU numbers. For example, if a worker has 2 GPUs,
    is set to ``[0, 1]``. If a worker is configured with subset of GPUs, only
    that subset is listed here e.g., ``[1]``.
    """
    RAM: List[int] = []
    """
    List of available 8GB slots. For example, if a worker has 32GB of RAM,
    is set to ``[0, 1, 2, 3]``.
    """


class FixedResourceSlots(BaseModel):
    """
    Available worker node compute resources that only indicate presence, not the
    amount that may be allocated. (e.g., "SSD is available or not available")
    """

    SSD: bool = False
    """
    Whether this target thas an SSD
    """


class ResourceSpec(BaseModel):
    """
    Job resource requirements. Used to allocate compute resources for a job
    at queue time.
    """

    cpu: int = 1
    """
    Number of required CPU cores
    """
    gpu: int = 0
    """
    Number of required GPUs
    """
    ram: int = 1
    """
    Number of 8GiB RAM slots
    """
    ssd: bool = False
    """
    Whether an SSD is required for temporary storage.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from .models import (
    api_request,
    api_response,
    asset,
    auth,
    benchmarks,
    config,
    diagnostics,
    event,
    exposure,
    external,
    file_browser,
    gpu,
    instance,
    job,
    job_register,
    job_spec,
    license,
    notification,
    params,
    preview,
    project,
    resource,
    scheduler_lane,
    scheduler_target,
    services,
    session,
    session_config_profile,
    session_params,
    session_spec,
    signature,
    tag,
    user,
    when,
    workspace,
)
from .registry import register_model_module

register_model_module(job, include_literals=True)
register_model_module(scheduler_target, include_literals=True)
register_model_module(resource, include_literals=True)
register_model_module(gpu, include_literals=True)
register_model_module(api_request, include_literals=True)
register_model_module(session, include_literals=True)
register_model_module(user, include_literals=True)
register_model_module(file_browser, include_literals=True)
register_model_module(job_spec, include_literals=True)
register_model_module(exposure, include_literals=True)
register_model_module(event, include_literals=True)
register_model_module(preview, include_literals=True)
register_model_module(session_params, include_literals=True)
register_model_module(external, include_literals=True)
register_model_module(project, include_literals=True)
register_model_module(api_response, include_literals=True)
register_model_module(asset, include_literals=True)
register_model_module(signature, include_literals=True)
register_model_module(instance, include_literals=True)
register_model_module(workspace, include_literals=True)
register_model_module(job_register, include_literals=True)
register_model_module(when, include_literals=True)
register_model_module(params, include_literals=True)
register_model_module(license, include_literals=True)
register_model_module(services, include_literals=True)
register_model_module(notification, include_literals=True)
register_model_module(benchmarks, include_literals=True)
register_model_module(session_config_profile, include_literals=True)
register_model_module(diagnostics, include_literals=True)
register_model_module(scheduler_lane, include_literals=True)
register_model_module(session_spec, include_literals=True)
register_model_module(config, include_literals=True)
register_model_module(tag, include_literals=True)
register_model_module(auth, include_literals=True)



from io import BytesIO

import numpy as n
import pytest

import cryosparc.dataset as ds

from .conftest import Dataset


@pytest.fixture
def dset(big_dset: Dataset):
    """Copy of big_set that can be mutated. big_dset should *NOT* but mutated
    because it's shared between tests!!!"""
    return big_dset.copy()


@pytest.fixture
def fields():
    return [
        ("uid", "<u8"),
        ("ctf/type", "|O"),
        ("ctf/exp_group_id", "<u4"),
        ("ctf/accel_kv", "<f4"),
        ("ctf/cs_mm", "<f4"),
        ("ctf/amp_contrast", "<f4"),
        ("ctf/df1_A", "<f4"),
        ("ctf/df2_A", "<f4"),
        ("ctf/df_angle_rad", "<f4"),
        ("ctf/phase_shift_rad", "<f4"),
        ("ctf/scale", "<f4"),
        ("ctf/scale_const", "<f4"),
        ("ctf/shift_A", "<f4", (2,)),
        ("ctf/tilt_A", "<f4", (2,)),
        ("ctf/trefoil_A", "<f4", (2,)),
        ("ctf/tetra_A", "<f4", (4,)),
        ("ctf/anisomag", "<f4", (4,)),
        ("ctf/bfactor", "<f4"),
        ("location/micrograph_uid", "<u8"),
        ("location/exp_group_id", "<u4"),
        ("location/micrograph_path", "|O"),
        ("location/micrograph_shape", "<u4", (2,)),
        ("location/center_x_frac", "<f4"),
        ("location/center_y_frac", "<f4"),
        ("pick_stats/ncc_score", "<f4"),
        ("pick_stats/power", "<f4"),
        ("pick_stats/template_idx", "<u4"),
        ("pick_stats/angle_rad", "<f4"),
    ]


@pytest.fixture
def dset_cstrs(big_dset: Dataset):
    return big_dset.to_cstrs(copy=True)


def test_len(big_dset: Dataset):
    assert len(big_dset) == 1961726


def test_fields(big_dset: Dataset):
    fields = big_dset.fields()
    assert len(fields) == 28
    # assert len(fields) == 26
    assert fields[0] == "uid"


def test_fields_no_uid(big_dset: Dataset):
    fields = big_dset.fields(exclude_uid=True)
    assert len(fields) == 27
    # assert len(fields) == 25
    assert fields[0] != "uid"


def test_dtypes(big_dset: Dataset):
    assert len(big_dset.descr()) == 28
    # assert len(big_dset.descr()) == 26


def test_contains(big_dset: Dataset):
    assert "location/micrograph_uid" in big_dset


def test_add_fields(dset: Dataset):
    dset.add_fields(["foo", "baz"], ["u8", "f4"])
    assert len(dset.fields()) == 30
    # assert len(dset.fields()) == 28


def test_rename_fields(dset: Dataset):
    renamed_fields = [f if f == "uid" else "{}__RENAMED".format(f) for f in dset.fields()]
    dset.rename_fields("{}__RENAMED".format)
    assert dset.fields() == renamed_fields


def test_get_items(benchmark, dset: Dataset):
    @benchmark
    def _():
        dset._rows = None
        assert len(dset.rows()) == 1961726


def test_get_items_to_list(benchmark, dset, fields):
    @benchmark
    def _():
        dset._rows = None
        items = dset.rows()
        first = items[0]
        lst = first.to_list()
        assert len(lst) == len(fields)
        assert any(lst)


def test_get_items_to_dict(benchmark, dset: Dataset, fields):
    @benchmark
    def _():
        dset._rows = None
        items = dset.rows()
        first = items[0]
        d = dict(first)
        assert set(d.keys()) == {f[0] for f in fields}
        assert any(d.values())


def test_get_items_to_item_dict(benchmark, dset: Dataset, fields):
    @benchmark
    def _():
        dset._rows = None
        first = dset.rows()[0]
        item_d = first.to_dict()
        assert set(item_d.keys()) == {f[0] for f in fields}
        assert any(item_d.values())


def test_filter_fields_list(benchmark, big_dset: Dataset):
    @benchmark
    def _():
        dset = big_dset.copy()
        fields = [
            "pick_stats/ncc_score",
            "pick_stats/power",
            "pick_stats/template_idx",
            "pick_stats/angle_rad",
        ]
        dset.filter_fields(fields)
        assert dset.fields() == ["uid"] + fields


def test_filter_prefixes(benchmark, big_dset: Dataset):
    @benchmark
    def _():
        dset = big_dset.copy()
        filtered = dset.filter_fields(lambda n: not n.startswith("pick_stats"))
        assert len(filtered.fields()) == 24
        # assert len(filtered.fields()) == 22


def test_copy_fields(benchmark, big_dset: Dataset):
    @benchmark
    def _():
        dset = big_dset.copy()
        dset.copy_fields(["ctf/type", "ctf/exp_group_id"], ["foo", "bar"])
        # dset.copy_fields(["ctf/exp_group_id"], ["bar"])
        assert all(dset["ctf/type"] == dset["foo"])
        assert all(dset["ctf/exp_group_id"] == dset["bar"])


def test_append(benchmark, big_dset: Dataset, dset: Dataset):
    dset = dset.reassign_uids()
    new_dset = benchmark(dset.append, big_dset, assert_same_fields=True)
    assert len(new_dset) == len(big_dset) * 2


def test_append_union(benchmark, big_dset: Dataset, dset: Dataset):
    other = benchmark(dset.union, big_dset)
    assert len(other) == len(big_dset)


def test_append_many(benchmark, big_dset, dset: Dataset):
    empty = Dataset.allocate(0)
    other = dset.copy().reassign_uids()
    new_dset = benchmark(Dataset.append, dset, empty, other)
    assert len(new_dset) == len(big_dset) * 2


def test_append_many_union(benchmark, big_dset, dset: Dataset):
    empty = Dataset.allocate(0)
    other = dset.copy().reassign_uids()
    new_dset = benchmark(Dataset.union, dset, dset, empty, other)
    assert len(new_dset) == len(big_dset) * 2


def test_append_many_union_repeat_allowed(benchmark, big_dset, dset: Dataset):
    empty = Dataset.allocate(0)
    other = dset.copy().reassign_uids()
    new_dset = benchmark(Dataset.append, dset, dset, empty, other, repeat_allowed=True)
    assert len(new_dset) == len(big_dset) * 3


def test_append_many_simple(benchmark, big_dset, dset: Dataset):
    empty = Dataset.allocate(0)
    other = dset.copy().reassign_uids()
    new_dset = benchmark(Dataset.append, dset, empty, other, assert_same_fields=True)
    assert len(new_dset) == len(big_dset) * 2


def test_append_many_simple_interlace(benchmark, big_dset, dset: Dataset):
    other = dset.copy()
    other.reassign_uids()

    new_dset = benchmark(Dataset.interlace, dset, other)
    assert len(new_dset) == len(big_dset) * 2
    assert new_dset["uid"][0] == dset["uid"][0]
    assert new_dset["uid"][1] == other["uid"][0]


def test_append_replace(benchmark, big_dset: Dataset, dset: Dataset):
    other = Dataset.allocate(size=10000, fields=dset.descr())
    new_dset = benchmark(dset.replace, {}, other)
    assert len(new_dset) == len(big_dset) + len(other)
    assert new_dset["uid"][-1] == other["uid"][-1]


def test_append_replace_unique(benchmark, big_dset: Dataset, dset: Dataset):
    other = Dataset.allocate(size=10000, fields=dset.descr())
    new_dset = benchmark(dset.replace, {}, other, assume_disjoint=True, assume_unique=True)
    assert len(new_dset) == len(big_dset) + len(other)
    assert new_dset["uid"][-1] == other["uid"][-1]


def test_append_replace_empty(benchmark, big_dset, dset: Dataset):
    other = Dataset.allocate(0, fields=dset.descr())
    new_dset = benchmark(dset.replace, {}, other)
    assert len(new_dset) == len(big_dset)
    assert new_dset == big_dset


def test_append_replace_empty_query(benchmark, big_dset: Dataset, dset: Dataset):
    other = Dataset.allocate(0, fields=dset.descr())
    new_dset = benchmark(dset.replace, {"location/micrograph_uid": 6655121610611186569}, other)
    assert len(new_dset) == len(big_dset) - 1191


def test_append_replace_query(benchmark, big_dset: Dataset, dset: Dataset):
    other = Dataset.allocate(size=10000, fields=dset.descr())
    new_dset = benchmark(dset.replace, {"location/micrograph_uid": 6655121610611186569}, other)
    assert len(new_dset) == len(big_dset) + len(other) - 1191


def test_append_replace_query_unique(benchmark, big_dset: Dataset, dset: Dataset):
    other = Dataset.allocate(size=10000, fields=dset.descr())
    new_dset = benchmark(
        dset.replace, {"location/micrograph_uid": 6655121610611186569}, other, assume_disjoint=True, assume_unique=True
    )
    assert len(new_dset) == len(big_dset) + len(other) - 1191


def test_append_replace_many(benchmark, big_dset: Dataset, dset: Dataset):
    other1 = Dataset.allocate(size=5000, fields=dset.descr())
    other2 = Dataset.allocate(size=5000, fields=dset.descr())
    new_dset = benchmark(dset.replace, {}, other1, other2)
    assert len(new_dset) == len(big_dset) + len(other1) + len(other2)
    assert new_dset["uid"][-1] == other2["uid"][-1]


def test_append_replace_many_unique(benchmark, big_dset: Dataset, dset: Dataset):
    other1 = Dataset.allocate(size=5000, fields=dset.descr())
    other2 = Dataset.allocate(size=5000, fields=dset.descr())
    new_dset = benchmark(dset.replace, {}, other1, other2, assume_disjoint=True, assume_unique=True)
    assert len(new_dset) == len(big_dset) + len(other1) + len(other2)
    assert new_dset["uid"][-1] == other2["uid"][-1]


def test_append_replace_many_query(benchmark, big_dset: Dataset, dset: Dataset):
    other1 = Dataset.allocate(size=5000, fields=dset.descr())
    other2 = Dataset.allocate(size=5000, fields=dset.descr())
    new_dset = benchmark(
        dset.replace,
        {"location/micrograph_uid": [2539634023577218663, 6655121610611186569]},
        other1,
        other2,
        assume_disjoint=True,
        assume_unique=True,
    )
    assert len(new_dset) == len(big_dset) + len(other1) + len(other2) - 1191 - 1210


def test_innerjoin_two(benchmark, dset: Dataset):
    other = dset.slice(500000, 1500000)
    expected = other
    joined = benchmark(dset.innerjoin, other.shuffle())
    assert joined == expected


def test_innerjoin_many(benchmark, dset: Dataset):
    other1 = dset.slice(500_000, 1_250_000)
    other2 = dset.slice(750_000, 1_500_000).shuffle()
    expected = dset.slice(750_000, 1_250_000)
    new_dset = benchmark(Dataset.innerjoin, dset, other1, other2)
    assert new_dset == expected


def test_subset_idxs(benchmark, dset: Dataset):
    new_dset = benchmark(dset.take, list(range(0, 1_500_000, 2)))  # Even entries up to 1.5 million
    assert len(new_dset) == 750_000


def test_subset_mask(benchmark, big_dset, dset: Dataset):
    # Even entries up to 1.5 million
    mask = n.array([i < 1500000 and i % 2 == 0 for i in range(len(dset))])
    new_dset = benchmark(dset.mask, mask)
    assert len(dset) == len(big_dset), "Should not mutate original dset"
    assert len(new_dset) == 750_000


def test_subset_query(benchmark, big_dset, dset: Dataset):
    new_dset = benchmark(dset.query, lambda item: item["location/micrograph_uid"] == 6655121610611186569)
    assert len(dset) == len(big_dset), "Should not mutate original dset"
    assert len(new_dset) == 1191


def test_subset_simple_query_1(benchmark, dset: Dataset):
    new_dset = benchmark(dset.query, {"location/micrograph_uid": 6655121610611186569})
    assert len(new_dset) == 1191


def test_subset_simple_query_2(benchmark, dset: Dataset):
    new_dset = benchmark(dset.query, {"location/micrograph_path": "J3/imported/18jam15a_0008_ali_DW.mrc"})
    assert len(dset) > 0
    assert len(new_dset) < len(dset)


def test_subset_simple_query_3(benchmark, dset: Dataset):
    new_dset = benchmark(
        dset.query, {"uid": dset["uid"][1000:2000], "location/micrograph_path": "J3/imported/18jam15a_0008_ali_DW.mrc"}
    )
    assert len(new_dset) > 0
    assert len(new_dset) < len(dset)


def test_subset_simple_query_empty(benchmark, dset: Dataset):
    new_dset = benchmark(dset.query, {})
    assert len(new_dset) == len(dset)


def test_subset_simple_query_nomatch(benchmark, big_dset, dset: Dataset):
    new_dset = benchmark(dset.query, {"uid": [42]})
    assert len(dset) == len(big_dset), "Should not mutate original dset"
    assert new_dset != dset
    assert len(new_dset) == 0


def test_subset_split_by(benchmark, dset: Dataset):
    dsets = benchmark(dset.split_by, "location/micrograph_uid")
    assert len(dsets) == 1644
    assert len(dsets[2539634023577218663]) == 1210


def test_items_split_by(benchmark, dset: Dataset):
    rows_split = benchmark(dset.rows().split_by, "location/micrograph_uid")
    assert len(rows_split) == 1644
    assert len(rows_split[2539634023577218663]) == 1210


def test_copy(benchmark, dset: Dataset):
    new_dset = benchmark(dset.copy)
    assert id(new_dset) != id(dset)
    assert new_dset == dset
    assert len(new_dset) == len(dset)


def test_streaming_bytes(benchmark, dset: Dataset):
    stream = BytesIO()

    @benchmark
    def _():
        total_bytes = 0
        for dat in dset.stream(compression="lz4"):
            stream.write(dat)
            total_bytes += len(dat)
        stream.seek(0)
        assert total_bytes > 0

    assert stream.read(6) == ds.FORMAT_MAGIC_PREFIXES[ds.NEWEST_FORMAT]


def test_from_streaming_bytes(benchmark, big_dset: Dataset):
    stream = BytesIO()
    for dat in big_dset.stream(compression="lz4"):
        stream.write(dat)

    def load():
        stream.seek(0)
        result = Dataset.load(stream)
        return result

    result = benchmark(load)
    assert len(result) == len(big_dset)


def test_to_cstrs(benchmark, dset: Dataset):
    assert dset.rows()[0]  # check that this is ok
    result: Dataset = benchmark(dset.to_cstrs, copy=True)
    assert result["ctf/type"].dtype.type == n.uint64
    assert dset.rows()[0]


def test_to_pystrs(benchmark, dset_cstrs: Dataset):
    result: Dataset = benchmark(dset_cstrs.to_pystrs, copy=True)
    assert result["ctf/type"].dtype.type == n.object_


def test_inspect(big_dset_path, fields):
    result = Dataset.inspect(big_dset_path)
    assert result["length"] == 1961726
    assert result["dtype"] == fields
    assert result["compression"] is None
    assert result["compressed_fields"] == []


def test_load(benchmark, big_dset_path, fields):
    result = benchmark(Dataset.load, big_dset_path)
    assert len(result) == 1961726
    assert result.descr() == fields


def test_load_prefixes(benchmark, big_dset_path, fields):
    keep_prefixes = ["location", "pick_stats"]
    expected_fields = [
        field
        for field in fields
        if field[0] == "uid" or field[0].startswith("location/") or field[0].startswith("pick_stats/")
    ]
    result = benchmark(Dataset.load, big_dset_path, prefixes=keep_prefixes)
    assert len(result) == 1961726
    assert result.descr() == expected_fields


def test_load_fields(benchmark, big_dset_path, fields):
    keep_fields = ["ctf/type", "ctf/shift_A", "location/micrograph_uid"]
    result = benchmark(Dataset.load, big_dset_path, fields=keep_fields)
    assert len(result) == 1961726
    assert result.descr() == [
        ("uid", "<u8"),
        ("ctf/type", "|O"),
        ("ctf/shift_A", "<f4", (2,)),
        ("location/micrograph_uid", "<u8"),
    ]


def test_load_prefixes_fields(benchmark, big_dset_path, fields):
    keep_prefixes = ["location", "pick_stats"]
    keep_fields = ["ctf/type", "ctf/shift_A", "location/micrograph_uid"]
    result = benchmark(Dataset.load, big_dset_path, prefixes=keep_prefixes, fields=keep_fields)
    expected_fields = [
        field
        for field in fields
        if field[0] == "uid"
        or field[0].startswith("location/")
        or field[0].startswith("pick_stats/")
        or field[0] in keep_fields
    ]
    assert len(result) == 1961726
    assert result.descr() == expected_fields



"""
Definitions for various error classes raised by cryosparc-tools functions
"""

import json
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from httpx import Response


class DatasetLoadError(Exception):
    """Exception type raised when a dataset cannot be loaded"""

    pass


class APIError(ValueError):
    """
    Raised by failed request to a CryoSPARC API server.
    """

    code: int
    data: Any = None

    def __init__(
        self,
        reason: str,
        *args: object,
        res: "Response",
        data: Any = None,  # must be JSON-encodable if provided
    ) -> None:
        msg = f"*** [API] ({res.request.method} {res.url}, code {res.status_code}) {reason}"
        super().__init__(msg, *args)
        self.code = res.status_code
        if data is not None:
            self.data = data
        elif res.headers.get("Content-Type") == "application/json":
            self.data = res.json()

    def __str__(self):
        s = super().__str__()
        if self.data:
            s += "\nResponse data:\n"
            s += json.dumps(self.data, indent=4)
        return s


class ExternalJobError(Exception):
    """
    Raised during external job lifecycle failures
    """

    pass



from enum import Enum
from typing import SupportsBytes

from numpy.typing import NDArray

__all__ = ["DsetType", "Stream", "Data"]

class MemoryView(SupportsBytes):  # Note: Supports buffer protocol.
    base: "Array"
    size: int
    itemsize: int
    nbytes: int
    ndim: int
    shape: tuple[int, ...]
    strides: tuple[int, ...]
    suboffsets: tuple[int, ...]
    T: "MemoryView"

    def copy(self) -> "MemoryView": ...
    def copy_fortran(self) -> "MemoryView": ...
    def is_c_contig(self) -> bool: ...
    def is_f_contig(self) -> bool: ...

class Array:
    memview: MemoryView

    def __len__(self) -> int: ...
    def __getitem__(self, key: int | slice) -> bytes: ...
    def __setitem__(self, key: int | slice, item: bytes): ...

class DsetType(int, Enum):
    T_F32 = ...
    T_F64 = ...
    T_C32 = ...
    T_C64 = ...
    T_I8 = ...
    T_I16 = ...
    T_I32 = ...
    T_I64 = ...
    T_U8 = ...
    T_U16 = ...
    T_U32 = ...
    T_U64 = ...
    T_STR = ...
    T_OBJ = ...

class Data:
    def __init__(self, other: "Data" | None = None) -> None: ...
    def innerjoin(self, key: str, other: "Data") -> "Data": ...
    def totalsz(self) -> int: ...
    def ncol(self) -> int: ...
    def nrow(self) -> int: ...
    def key(self, index: int) -> str: ...
    def type(self, field: str) -> DsetType: ...
    def has(self, field: str) -> bool: ...
    def addrows(self, num: int) -> None: ...
    def addcol_scalar(self, field: str, dtype: int) -> None: ...
    def addcol_array(self, field: str, dtype: int, shape: tuple[int, ...]) -> None: ...
    def getshp(self, colkey: str) -> tuple[int, ...]: ...
    def getbuf(self, colkey: str) -> Array | None: ...
    def getstr(self, col: str, index: int) -> bytes: ...
    def tocstrs(self, col: str) -> bool: ...
    def topystrs(self, col: str) -> bool: ...
    def stralloc(self, val: str) -> int: ...
    def dump(self) -> Array: ...
    def dumpstrheap(self) -> Array: ...
    def setstrheap(self, heap: bytes) -> None: ...
    def defrag(self, realloc_smaller: bool) -> None: ...
    def dumptxt(self, dump_data: bool = False) -> None: ...
    def handle(self) -> int: ...

class Stream:
    def __init__(self, data: Data) -> None: ...
    def cast_objs_to_strs(self) -> None: ...
    def stralloc_col(self, col: str) -> Array | None: ...
    def compress_col(self, col: str) -> Array: ...
    def compress_numpy(self, arr: NDArray) -> Array: ...
    def compress(self, arr: Array) -> Array: ...
    def decompress_col(self, col: str, data: bytes) -> Array: ...
    def decompress_numpy(self, data: bytes, arr: NDArray) -> Array: ...
    def decompress(self, data: bytes, outptr: int = 0) -> Array: ...



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import List, Literal, Optional

from pydantic import BaseModel


class BrowseFile(BaseModel):
    """
    Representation of a file or directory in the file browser.
    """

    file_name: str
    """
    Base name of file or directory.
    """
    base_path: str
    """
    Absolute path of the parent directory.
    """
    is_hidden: bool
    """
    """
    path_abs: str
    """
    """
    is_link: bool = False
    """
    """
    mtime: Optional[float] = None
    """
    """
    size: Optional[int] = None
    """
    """
    type: Optional[Literal["dir", "file"]] = None
    """
    """
    link_path: Optional[str] = None
    """
    """
    errmesg: Optional[str] = None
    """
    """


class BrowseFileResponse(BaseModel):
    """
    Response information when using the instance file browser API.
    """

    back_path: str
    """
    """
    files: List[BrowseFile]
    """
    """
    type: str
    """
    """


class FileBrowserPrefixes(BaseModel):
    """
    Configuration for allowed file browser prefixes and default directories.
    """

    allowed_prefixes: Optional[List[str]] = None
    """
    """
    import_dir_default: Optional[str] = None
    """
    """
    project_dir_default: Optional[str] = None
    """
    """



import base64
from datetime import datetime
from enum import Enum
from pathlib import PurePath
from typing import Any, Mapping

import numpy as n
from pydantic import BaseModel


def api_encode(obj: Any):
    """
    Recursively encode any object for transmission through the API.
    """
    if isinstance(obj, dict):
        return {k: api_encode(v) for k, v in obj}
    elif isinstance(obj, list):
        return [api_encode(v) for v in obj]
    else:
        return api_default(obj)


def api_default(obj: Any) -> Any:
    """
    json.dump "default" argument for sending objects over a JSON API. Ensures
    that special non-JSON types such as Path are NDArray are encoded correctly.
    """
    if isinstance(obj, n.floating):
        if n.isnan(obj):
            return float(0)
        elif n.isposinf(obj):
            return float("inf")
        elif n.isneginf(obj):
            return float("-inf")
        return float(obj)
    elif isinstance(obj, n.integer):
        return int(obj)
    elif isinstance(obj, n.ndarray):
        return ndarray_to_json(obj)
    elif isinstance(obj, bytes):
        return binary_to_json(obj)
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, PurePath):
        return str(obj)
    elif isinstance(obj, BaseModel):
        return obj.model_dump(mode="json", exclude_none=True)
    else:
        return obj


def api_object_hook(dct: Mapping[str, Any]):
    """
    json.dump "object_hook" argument for receiving JSON from an API request.
    Ensures that special objects that are actually encoded numpy arrays or bytes
    are decoded as such.
    """
    if "$ndarray" in dct:
        return ndarray_from_json(dct)
    elif "$binary" in dct:
        return binary_from_json(dct)
    else:
        return dct  # pydantic will take care of everything else


def binary_to_json(binary: bytes):
    """
    Encode bytes as a JSON-serializeable object
    """
    return {"$binary": {"base64": base64.b64encode(binary).decode()}}


def binary_from_json(dct: Mapping[str, Any]) -> bytes:
    if "base64" not in dct["$binary"] or not isinstance(b64 := dct["$binary"]["base64"], str):
        raise TypeError(f"$binary base64 must be a string: {dct}")
    return base64.b64decode(b64.encode())


def ndarray_to_json(arr: n.ndarray):
    """
    Encode a numpy array a JSON-serializeable object.
    """
    return {
        "$ndarray": {
            "base64": base64.b64encode(arr.data).decode(),
            "dtype": str(arr.dtype),
            "shape": arr.shape,
        }
    }


def ndarray_from_json(dct: Mapping[str, Any]):
    """
    Decode a serialized numpy array.
    """
    data = base64.b64decode(dct["$ndarray"]["base64"])
    return n.frombuffer(data, dct["$ndarray"]["dtype"]).reshape(dct["$ndarray"]["shape"])



"""
Defines the Job and External job classes for accessing CryoSPARC jobs.
"""

import re
import traceback
import warnings
from contextlib import contextmanager
from io import BytesIO
from pathlib import PurePath, PurePosixPath
from time import sleep, time
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
    overload,
)

from ..dataset import DEFAULT_FORMAT, Dataset
from ..errors import ExternalJobError
from ..models.asset import GridFSAsset, GridFSFile
from ..models.job import Job, JobStatus
from ..models.job_register import JobRegisterJobSpec
from ..models.job_spec import Input, InputSpec, Output, OutputSpec, Params
from ..spec import (
    ASSET_CONTENT_TYPES,
    IMAGE_CONTENT_TYPES,
    TEXT_CONTENT_TYPES,
    AssetFormat,
    Datatype,
    ImageFormat,
    LoadableSlots,
    SlotSpec,
    TextFormat,
)
from ..stream import Stream
from ..util import PurePosixPathProperty, first, print_table
from . import Controller, as_input_slot, as_output_slot

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray

    from ..tools import CryoSPARC


GROUP_NAME_PATTERN = r"^[A-Za-z][0-9A-Za-z_]*$"
"""
Input and output result groups may only contain, letters, numbers and underscores.
"""

LogLevel = Literal["text", "warning", "error"]
"""
Severity level for job event logs.
"""

FileOrFigure = Union[str, PurePath, IO[bytes], Any]
"""A file path, a file-like object, or a matplotlib figure."""


class JobController(Controller[Job]):
    """
    Accessor class to a job in CryoSPARC with ability to load inputs and
    outputs, add to job log, download job files. Should be created with
    :py:meth:`cs.find_job() <cryosparc.tools.CryoSPARC.find_job>` or
    :py:meth:`project.find_job() <cryosparc.controllers.project.ProjectController.find_job>`.

    Arguments:
        job (tuple[str, str] | Job): either _(Project UID, Job UID)_ tuple or
            Job model, e.g. ``("P3", "J42")``

    Attributes:
        model (Job): All job data from the CryoSPARC database.
            Contents may change over time, use :py:meth:`refresh` to update.

    Examples:

        Find an existing job.

        >>> cs = CryoSPARC("http://localhost:61000")
        >>> job = cs.find_job("P3", "J42")
        >>> job.status
        "building"

        Queue a job.

        >>> job.queue("worker_lane")
        >>> job.status
        "queued"

        Create a 3-class ab-initio job connected to existing particles.

        >>> job = cs.create_job("P3", "W1", "homo_abinit"
        ...     connections={"particles": ("J20", "particles_selected")}
        ...     params={"abinit_K": 3}
        ... )
        >>> job.queue()
        >>> job.status
        "queued"
    """

    uid: str
    """
    Job unique ID, e.g., "J42"
    """
    project_uid: str
    """
    Project unique ID, e.g., "P3"
    """

    _events: Dict[str, str]
    """
    Named event logs. Key can be user-provided name in log() method, or ID if
    name not provided. If both name and ID are used, can have two keys with
    the same value.

    :meta private:
    """

    def __init__(self, cs: "CryoSPARC", job: Union[Tuple[str, str], Job]) -> None:
        self.cs = cs
        if isinstance(job, tuple):
            self.project_uid, self.uid = job
            self.refresh()
        else:
            self.project_uid = job.project_uid
            self.uid = job.uid
            self.model = job
        self._events = {}

    @property
    def type(self) -> str:
        """Job type key"""
        return self.model.spec.type

    @property
    def title(self) -> str:
        """Job title"""
        return self.model.title

    @property
    def desc(self) -> str:
        """Job description"""
        return self.model.description

    @property
    def status(self) -> JobStatus:
        """Job scheduling status."""
        return self.model.status

    @property
    def dir(self) -> PurePosixPath:
        """Full path to the job directory."""
        return PurePosixPathProperty(self.cs.api.jobs.get_directory(self.project_uid, self.uid))

    @property
    def params(self) -> Params:
        """
        Job parameter values object.

        Example:
            >>> cs = CryoSPARC(...)
            >>> job = cs.find_job("P3", "J42")
            >>> print(job.type)
            "homo_abinit"
            >>> print(job.params.abinit_K)
            3
        """
        return self.model.spec.params

    @property
    def inputs(self) -> Dict[str, Input]:
        """Input connection details."""
        return self.model.spec.inputs.root

    @property
    def outputs(self) -> Dict[str, Output]:
        """Input result details."""
        return self.model.spec.outputs.root

    @property
    def full_spec(self) -> JobRegisterJobSpec:
        """
        The full specification for job inputs, outputs and parameters, as
        defined in the job register.
        """
        spec = first(spec for spec in self.cs.job_register.specs if spec.type == self.type)
        if not spec:
            raise RuntimeError(f"Could not find job specification for type {type}")
        return spec

    def refresh(self):
        """
        Reload this job from the CryoSPARC database.

        Returns:
            JobController: self
        """
        self.model = self.cs.api.jobs.find_one(self.project_uid, self.uid)
        return self

    def queue(
        self,
        lane: Optional[str] = None,
        hostname: Optional[str] = None,
        gpus: List[int] = [],
        cluster_vars: Dict[str, Any] = {},
    ):
        """
        Queue a job to a target lane. Available lanes may be queried with
        `:py:meth:`cs.get_lanes() <cryosparc.tools.CryoSPARC.get_lanes>`.

        Optionally specify a hostname for a node or cluster in the given lane.
        Optionally specify specific GPUs indexes to use for computation.

        Available hostnames for a given lane may be queried with
        `:py:meth:`cs.get_targets() <cryosparc.tools.CryoSPARC.get_targets>`.

        Args:
            lane (str, optional): Configuried compute lane to queue to. Leave
                unspecified to run directly on the master or current
                workstation. Defaults to None.
            hostname (str, optional): Specific hostname in compute lane, if more
                than one is available. Defaults to None.
            gpus (list[int], optional): GPUs to queue to. If specified, must
                have as many GPUs as required in job parameters. Leave
                unspecified to use first available GPU(s). Defaults to [].
            cluster_vars (dict[str, Any], optional): Specify custom cluster
                variables when queuing to a cluster. Keys are variable names.
                Defaults to False.

        Examples:

            Queue a job to lane named "worker":

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> job.status
            "building"
            >>> job.queue("worker")
            >>> job.status
            "queued"
        """
        if cluster_vars:
            self.cs.api.jobs.set_cluster_custom_vars(self.project_uid, self.uid, cluster_vars)
        self.model = self.cs.api.jobs.enqueue(self.project_uid, self.uid, lane=lane, hostname=hostname, gpus=gpus)

    def kill(self):
        """
        Kill this job.
        """
        self.model = self.cs.api.jobs.kill(self.project_uid, self.uid)

    def wait_for_status(self, status: Union[JobStatus, Iterable[JobStatus]], *, timeout: Optional[int] = None) -> str:
        """
        Wait for a job's status to reach the specified value. Must be one of
        the following:

        - 'building'
        - 'queued'
        - 'launched'
        - 'started'
        - 'running'
        - 'waiting'
        - 'completed'
        - 'killed'
        - 'failed'

        Args:
            status (str | set[str]): Specific status or set of statuses to wait
                for. If a set of statuses is specified, waits util job reaches
                any of the specified statuses.
            timeout (int, optional): If specified, wait at most this many
                seconds. Once timeout is reached, returns current status.
                Defaults to None.

        Returns:
            str: current job status
        """
        statuses = {status} if isinstance(status, str) else set(status)
        tic = time()
        while self.refresh().status not in statuses:
            if timeout is not None and time() - tic > timeout:
                break
            sleep(5)
        return self.status

    def wait_for_done(self, *, error_on_incomplete: bool = False, timeout: Optional[int] = None) -> str:
        """
        Wait until a job reaches status "completed", "killed" or "failed".

        Args:
            error_on_incomplete (bool, optional): If True, raises an assertion
                error when job finishes with status other than "completed" or
                timeout is reached. Defaults to False.
            timeout (int, optional): If specified, wait at most this many
                seconds. Once timeout is reached, returns current status or
                fails if ``error_on_incomplete`` is ``True``. Defaults to None.
        """
        status = self.wait_for_status({"completed", "killed", "failed"}, timeout=timeout)
        assert not error_on_incomplete or status == "completed", (
            f"Job {self.project_uid}-{self.uid} did not complete (status {status})"
        )
        return status

    def interact(self, action: str, body: Any = {}, *, timeout: int = 10, refresh: bool = False) -> Any:
        """
        Call an interactive action on a waiting interactive job. The possible
        actions and expected body depends on the job type.

        Args:
            action (str): Interactive endpoint to call.
            body (any): Body parameters for the interactive endpoint. Must be
                JSON-encodable.
            timeout (int, optional): Maximum time to wait for the action to
                complete, in seconds. Defaults to 10.
            refresh (bool, optional): If True, refresh the job document after
                posting. Defaults to False.
        """
        result: Any = self.cs.api.jobs.interactive_post(
            self.project_uid, self.uid, body=body, endpoint=action, timeout=timeout
        )
        if refresh:
            self.refresh()
        return result

    def clear(self):
        """
        Clear this job and reset to building status.
        """
        self.model = self.cs.api.jobs.clear(self.project_uid, self.uid)

    def set_title(self, title: str):
        """
        Set the job title.

        Args:
            title (str): New job title
        """
        self.model = self.cs.api.jobs.set_title(self.project_uid, self.uid, title=title)

    def set_description(self, desc: str):
        """
        Set the job description. May include Markdown formatting.

        Args:
            desc (str): New job description
        """
        self.model = self.cs.api.jobs.set_description(self.project_uid, self.uid, description=desc)

    def set_param(self, name: str, value: Any, **kwargs) -> bool:
        """
        Set the given param name on the current job to the given value. Only
        works if the job is in "building" status.

        Args:
            name (str): Param name, as defined in the job document's ``params_base``.
            value (any): Target parameter value.

        Returns:
            bool: False if the job encountered a build error.

        Examples:

            Set the number of GPUs used by a supported job

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> job.set_param("compute_num_gpus", 4)
            True
        """
        if "refresh" in kwargs:
            warnings.warn("refresh argument no longer applies", DeprecationWarning, stacklevel=2)
        self.model = self.cs.api.jobs.set_param(self.project_uid, self.uid, name, value=value)
        return True

    def connect(self, target_input: str, source_job_uid: str, source_output: str, **kwargs) -> bool:
        """
        Connect the given input for this job to an output with given job UID and
        name.

        Args:
            target_input (str): Input name to connect into. Will be created if
                not specified.
            source_job_uid (str): Job UID to connect from, e.g., "J42"
            source_output (str): Job output name to connect from , e.g.,
                "particles"

        Returns:
            bool: False if the job encountered a build error.

        Examples:

            Connect J3 to CTF-corrected micrographs from J2's ``micrographs``
            output.

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_job("J3")
            >>> job.connect("input_micrographs", "J2", "micrographs")

        """
        if "refresh" in kwargs:
            warnings.warn("refresh argument no longer applies", DeprecationWarning, stacklevel=2)
        if source_job_uid == self.uid:
            raise ValueError(f"Cannot connect job {self.uid} to itself")
        self.model = self.cs.api.jobs.connect(
            self.project_uid, self.uid, target_input, source_job_uid=source_job_uid, source_output_name=source_output
        )
        return True

    def connect_result(
        self,
        target_input: str,
        connection_idx: int,
        slot: str,
        source_job_uid: str,
        source_output: str,
        source_result: str,
        source_version: Union[int, Literal["F"]] = "F",
    ):
        """
        Connect a low-level input result slot with a result from another job.

        Args:
            target_input (str): Input name to connect into, e.g., "particles"
            connection_idx (int): Connection index to connect into, use 0 for
                the job's first connection on that input, 1 for the second, etc.
            slot (str): Input slot name to connect into, e.g., "location"
            source_job_uid (str): Job UID to connect from, e.g., "J42"
            source_output (str): Job output name to connect from , e.g.,
                "particles_selected"
            source_result (str): Result name to connect from, e.g., "location"

        Returns:
            bool: False if the job encountered a build error.

        Examples:

            Connect J3 to the first connection of J2's ``particles`` input.
            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_job("J3")
            >>> job.connect_result("particles", 0, "location", "J2", "particles_selected", "location")
        """
        assert source_job_uid != self.uid, f"Cannot connect job {self.uid} to itself"
        self.model = self.cs.api.jobs.connect_result(
            self.project_uid,
            self.uid,
            target_input,
            connection_idx,
            slot,
            source_job_uid=source_job_uid,
            source_output_name=source_output,
            source_result_name=source_result,
            source_result_version=source_version,
        )
        return True

    def disconnect(self, target_input: str, connection_idx: Optional[int] = None, **kwargs):
        """
        Clear the given job input group.

        Args:
            target_input (str): Name of input to disconnect
            connection_idx (int, optional): Connection index to clear.
                Set to 0 to clear the first connection, 1 for the second, etc.
                If unspecified, clears all connections. Defaults to None.
        """
        if "refresh" in kwargs:
            warnings.warn("refresh argument no longer applies", DeprecationWarning, stacklevel=2)

        if connection_idx is None:  # Clear all input connections
            self.model = self.cs.api.jobs.disconnect_all(self.project_uid, self.uid, target_input)
        else:
            self.model = self.cs.api.jobs.disconnect(self.project_uid, self.uid, target_input, connection_idx)

    def disconnect_result(self, target_input: str, connection_idx: int, slot: str):
        """
        Clear the job's given input result slot.

        Args:
            target_input (str): Name of input to disconnect
            connection_idx (int): Connection index to modify. Set to 0 for the
                first connection, 1 for the second, etc.
            slot (str): Input slot name to disconnect, e.g., "location"

        Returns:
            bool: False if the job encountered a build error.
        """
        self.model = self.cs.api.jobs.disconnect_result(self.project_uid, self.uid, target_input, connection_idx, slot)
        return True

    def load_input(self, name: str, slots: LoadableSlots = "all"):
        """
        Load the dataset connected to the job's input with the given name.

        Args:
            name (str): Input to load
            slots (Literal["default", "passthrough", "all"] | list[str], optional):
                List of specific slots to load, such as ``movie_blob`` or
                ``locations``, or all slots if not specified (including
                passthrough). May also specify as keyword. Defaults to
                "all".

        Raises:
            TypeError: If the job doesn't have the given input or the dataset
                cannot be loaded.

        Returns:
            Dataset: Loaded dataset
        """
        return self.cs.api.jobs.load_input(self.project_uid, self.uid, name, slots=slots)

    def load_output(self, name: str, slots: LoadableSlots = "all", version: Union[int, Literal["F"]] = "F"):
        """
        Load the dataset for the job's output with the given name.

        Args:
            name (str): Output to load
            slots (Literal["default", "passthrough", "all"] | list[str], optional):
                List of specific slots to load, such as ``movie_blob`` or
                ``locations``, or all slots if not specified (including
                passthrough). May also specify as keyword. Defaults to
                "all".
            version (int | Literal["F"], optional): Specific output version to
                load. Use this to load the output at different stages of
                processing. Leave unspecified to load final verion. Defaults to
                "F"

        Raises:
            TypeError: If job does not have any results for the given output

        Returns:
            Dataset: Loaded dataset
        """
        return self.cs.api.jobs.load_output(self.project_uid, self.uid, name, slots=slots, version=version)

    @overload
    def log(self, text: str, *, level: LogLevel = ...) -> str: ...
    @overload
    def log(self, text: str, *, level: LogLevel = ..., name: str) -> str: ...
    @overload
    def log(self, text: str, *, level: LogLevel = ..., id: str) -> str: ...
    def log(self, text: str, *, level: LogLevel = "text", name: Optional[str] = None, id: Optional[str] = None) -> str:
        """
        Append to a job's event log. Update an existing log by providing a name
        or ID.

        Args:
            text (str): Text to log
            level (str, optional): Log level ("text", "warning" or "error").
                Defaults to "text".
            name (str, optional): Event name. If called multiple times with the
                same name, updates that event instead of creating a new one.
                Named events are reset when logging a checkpoint. Cannot be
                provided with id. Defaults to None.
            id (str, optional): Update a previously-created event log by its ID.
                Cannot be provided with name. Defaults to None.

        Example:

            Log a warning message to the job log.
            >>> job.log("This is a warning", level="warning")

            Show a live progress bar in the job log.
            >>> for pct in range(1, 10):
            ...     # example log: "Progress: [#####-----] 50%"
            ...     job.log(f"Progress: [{'#' * pct}{'-' * (10 - pct)}] {pct * 10}%", name="progress")
            ...     sleep(1)
            ...
            >>> job.log("Done!")

            Update an existing log event by ID.
            >>> event_id = job.log("Starting job processing...")
            >>> # do some processing...
            >>> job.log("Finished processing", id=event_id)

        Returns:
            str: Created log event ID
        """
        existing_id = id
        if name and name in self._events:
            existing_id = self._events[name]

        if existing_id:
            event = self.cs.api.jobs.update_event_log(self.project_uid, self.uid, existing_id, text, type=level)
        else:
            event = self.cs.api.jobs.add_event_log(self.project_uid, self.uid, text, type=level)

        if name:
            self._events[name] = event.id
        return event.id

    def log_checkpoint(self, meta: dict = {}):
        """
        Append a checkpoint to the job's event log. Also resets named events.

        Args:
            meta (dict, optional): Additional meta information. Defaults to {}.

        Returns:
            str: Created checkpoint event ID
        """
        event = self.cs.api.jobs.add_checkpoint(self.project_uid, self.uid, meta)
        self._events = {}
        return event.id

    def log_plot(
        self,
        figure: FileOrFigure,
        text: str,
        formats: Iterable[ImageFormat] = ["png", "pdf"],
        raw_data: Union[str, bytes, None] = None,
        raw_data_file: Union[str, PurePath, IO[bytes], None] = None,
        raw_data_format: Optional[TextFormat] = None,
        flags: List[str] = ["plots"],
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ):
        """
        Add a log line with the given figure.

        ``figure`` must be one of the following

        - Path to an existing image file in PNG, JPEG, GIF, SVG or PDF format
        - A file handle-like object with the binary data of an image
        - A matplotlib plot

        If a matplotlib figure is specified, Uploads the plots in ``png`` and
        ``pdf`` formats. Override the ``formats`` argument with
        ``formats=['<format1>', '<format2>', ...]`` to save in different image
        formats.

        If a text-version of the given plot is available (e.g., in ``csv``
        format), specify ``raw_data`` with the full contents or
        ``raw_data_file`` with a path or binary file handle pointing to the
        contents. Assumes file format from extension or ``raw_data_format``.
        Defaults to ``"txt"`` if cannot be determined.

        Args:
            figure (str | Path | IO | Figure): Image file path, file handle or
                matplotlib figure instance
            text (str): Associated description for given figure
            formats (list[ImageFormat], optional): Image formats to save plot
                into. If a ``figure`` is a file handle, specify
                ``formats=['<format>']``, where ``<format>`` is a valid image
                extension such as ``png`` or ``pdf``. Assumes ``png`` if not
                specified. Defaults to ["png", "pdf"].
            raw_data (str | bytes, optional): Raw text data for associated plot,
                generally in CSV, XML or JSON format. Cannot be specified with
                ``raw_data_file``. Defaults to None.
            raw_data_file (str | Path | IO, optional): Path to raw text data.
                Cannot be specified with ``raw_data``. Defaults to None.
            raw_data_format (TextFormat, optional): Format for raw text data.
                Defaults to None.
            flags (list[str], optional): Flags to use for UI rendering.
                Generally should not be specified. Defaults to ["plots"].
            savefig_kw (dict, optional): If a matplotlib figure is specified
                optionally specify keyword arguments for the ``savefig`` method.
                Defaults to dict(bbox_inches="tight", pad_inches=0).

        Returns:
            str: Created log event ID
        """
        imgfiles = self._upload_plot(
            figure,
            name=text,
            formats=formats,
            raw_data=raw_data,
            raw_data_file=raw_data_file,
            raw_data_format=raw_data_format,
            savefig_kw=savefig_kw,
        )
        event = self.cs.api.jobs.add_image_log(self.project_uid, self.uid, imgfiles, text=text, flags=flags)
        return event.id

    def list_files(self, prefix: Union[str, PurePosixPath] = "", recursive: bool = False) -> List[str]:
        """
        Get a list of files inside the job directory.

        Args:
            prefix (str | Path, optional): Subdirectory inside job to list.
                Defaults to "".
            recursive (bool, optional): If True, lists files recursively.
                Defaults to False.

        Returns:
            list[str]: List of file paths relative to the job directory.
        """
        root = PurePosixPath(self.uid)
        files = self.cs.list_files(self.project_uid, prefix=root / prefix, recursive=recursive)
        return [str(PurePosixPath(f).relative_to(root)) for f in files]  # Strip leading "J#/"

    def download(self, path: Union[str, PurePosixPath]):
        """
        Initiate a download request for a file inside the job's directory. Use
        to get files from a remote CryoSPARC instance where the job directory
        is not available on the client file system.

        Args:
            path (str | Path): Name or path of file in job directory.

        Yields:
            HTTPResponse: Use a context manager to read the file from the
                request body.

        Examples:

            Download a job's metadata

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> with job.download("job.json") as res:
            >>>     job_data = json.loads(res.read())

        """
        path = PurePosixPath(self.uid) / path
        return self.cs.download(self.project_uid, path)

    def download_file(self, path: Union[str, PurePosixPath], target: Union[str, PurePath, IO[bytes]] = ""):
        """
        Download file from job directory to the given target path or writeable
        file handle.

        Args:
            path (str | Path): Name or path of file in job directory.
            target (str | Path | IO): Local file path, directory path or
                writeable file handle to write response data. If not specified,
                downloads to current working directory with same file name.
                Defaults to "".

        Returns:
            Path | IO: resulting target path or file handle.
        """
        path = PurePosixPath(self.uid) / path
        return self.cs.download_file(self.project_uid, path, target)

    def download_dataset(self, path: Union[str, PurePosixPath]):
        """
        Download a .cs dataset file from the given path in the job
        directory.

        Args:
            path (str | Path): Name or path of .cs file in job directory.

        Returns:
            Dataset: Loaded dataset instance
        """
        path = PurePosixPath(self.uid) / path
        return self.cs.download_dataset(self.project_uid, path)

    def download_mrc(self, path: Union[str, PurePosixPath]):
        """
        Download a .mrc file from the given relative path in the job directory.

        Args:
            path (str | Path): Name or path of .mrc file in job directory.

        Returns:
            tuple[Header, NDArray]: MRC file header and data as a numpy array
        """
        path = PurePosixPath(self.uid) / path
        return self.cs.download_mrc(self.project_uid, path)

    def list_assets(self) -> List[GridFSFile]:
        """
        Get a list of files available in the database for this job. Returns a
        list with details about the assets. Each entry is a dict with a ``_id``
        key which may be used to download the file with the ``download_asset``
        method.

        Returns:
            list[GridFSFile]: Asset details
        """
        return self.cs.list_assets(self.project_uid, self.uid)

    def download_asset(self, fileid: str, target: Union[str, PurePath, IO[bytes]]):
        """
        Download a job asset from the database with the given ID. Note that the
        file does not necessary have to belong to the current job.

        Args:
            fileid (str): GridFS file object ID
            target (str | Path | IO): Local file path or writeable file handle
                to write response data.

        Returns:
            str | Path | IO: resulting target path or file handle.

        """
        return self.cs.download_asset(fileid, target)

    def upload(
        self,
        target_path: Union[str, PurePosixPath],
        source: Union[str, bytes, PurePath, IO],
        *,
        overwrite: bool = False,
    ):
        """
        Upload the given file to the job directory at the given path. Fails if
        target already exists.

        Args:
            target_path (str | Path): Name or path of file to write in job
                directory.
            source (str | bytes | Path | IO): Local path or file handle to
                upload. May also specified as raw bytes.
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        target_path = PurePosixPath(self.uid) / target_path
        return self.cs.upload(self.project_uid, target_path, source, overwrite=overwrite)

    def _upload_asset(
        self,
        file: Union[str, PurePath, IO[bytes]],
        filename: Optional[str] = None,
        format: Optional[AssetFormat] = None,
    ) -> GridFSAsset:
        """
        Upload an image or text file to the current job. Specify either an image
        (PNG, JPG, GIF, PDF, SVG), text file (TXT, CSV, JSON, XML) or a binary
        IO object with data in one of those formats.

        If a binary IO object is specified, either a filename or file format
        must be specified.

        Do not use directly; use e.g., ``log_plot`` instead.
        """
        ext = None
        if format:
            ext = format
        elif filename:
            ext = filename.split(".")[-1].lower()
        elif isinstance(file, (str, PurePath)):
            file = PurePath(file)
            filename = file.name
            ext = file.suffix[1:].lower()
        else:
            raise ValueError("Must specify filename or format when saving binary asset")
        if ext not in ASSET_CONTENT_TYPES:
            raise ValueError(f"Invalid asset format {ext}")
        return self.cs.api.assets.upload(self.project_uid, self.uid, Stream.load(file), filename=filename, format=ext)

    def _upload_plot(
        self,
        figure: FileOrFigure,
        name: Optional[str] = None,
        formats: Iterable[ImageFormat] = ["png", "pdf"],
        raw_data: Union[str, bytes, None] = None,
        raw_data_file: Union[str, PurePath, IO[bytes], None] = None,
        raw_data_format: Optional[TextFormat] = None,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ) -> List[GridFSAsset]:
        """
        Upload the given figure. Returns a list of the created asset objects.
        Avoid using directly; use ``log_plot`` instead. See ``log_plot``
        additional details.
        """
        figdata = []
        basename = name or "figure"
        if hasattr(figure, "savefig"):  # matplotlib plot
            for fmt in formats:
                if fmt not in IMAGE_CONTENT_TYPES:
                    raise ValueError(f"Invalid figure format {fmt}")
                filename = f"{basename}.{fmt}"
                data = BytesIO()
                figure.savefig(data, format=fmt, **savefig_kw)  # type: ignore
                data.seek(0)
                figdata.append((data, filename, fmt))
        elif isinstance(figure, (str, PurePath)):  # file path; assume format from filename
            path = PurePath(figure)
            basename = path.stem
            fmt = path.suffix[1:].lower()
            if fmt not in IMAGE_CONTENT_TYPES:
                raise ValueError(f"Invalid figure format {fmt}")
            filename = f"{name or path.stem}.{fmt}"
            figdata.append((figure, filename, fmt))
        else:  # Binary IO
            fmt = first(iter(formats))
            if fmt not in IMAGE_CONTENT_TYPES:
                raise ValueError(f"Invalid or unspecified figure format {fmt}")
            filename = f"{basename}.{fmt}"
            figdata.append((figure, filename, fmt))

        raw_data_filename = f"{name}.{raw_data_format or 'txt'}" if name else None
        if raw_data and raw_data_file:
            raise ValueError("Ambiguous invocation; specify only one of raw_data or raw_data_file")
        elif isinstance(raw_data, str):
            raw_data_file = BytesIO(raw_data.encode())
        elif isinstance(raw_data, bytes):
            raw_data_file = BytesIO(raw_data)
        elif isinstance(raw_data_file, (str, PurePath)):
            raw_data_path = PurePath(raw_data_file)
            raw_data_filename = raw_data_path.name
            ext = raw_data_format or raw_data_filename.split(".")[-1]
            if ext not in TEXT_CONTENT_TYPES:
                raise ValueError(f"Invalid raw data filename {raw_data_file}")
            raw_data_format = ext

        assets = [self._upload_asset(data, filename, fmt) for data, filename, fmt in figdata]
        if raw_data_file:
            asset = self._upload_asset(raw_data_file, raw_data_filename, raw_data_format or "txt")
            assets.append(asset)

        return assets

    def upload_dataset(
        self,
        target_path: Union[str, PurePosixPath],
        dset: Dataset,
        *,
        format: int = DEFAULT_FORMAT,
        overwrite: bool = False,
    ):
        """
        Upload a dataset as a CS file into the job directory. Fails if target
        already exists.

        Args:
            target_path (str | Path): Name or path of dataset to save in the job
                directory. Should have a ``.cs`` extension.
            dset (Dataset): Dataset to save.
            format (int): Format to save in from ``cryosparc.dataset.*_FORMAT``,
                defaults to NUMPY_FORMAT)
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        target_path = PurePosixPath(self.uid) / target_path
        return self.cs.upload_dataset(self.project_uid, target_path, dset, format=format, overwrite=overwrite)

    def upload_mrc(
        self,
        target_path: Union[str, PurePosixPath],
        data: "NDArray",
        psize: float,
        *,
        overwrite: bool = False,
    ):
        """
        Upload a numpy 2D or 3D array to the job directory as an MRC file. Fails
        if target already exists.

        Args:
            target_path (str | Path): Name or path of MRC file to save in the
                job directory. Should have a ``.mrc`` extension.
            data (NDArray): Numpy array with MRC file data.
            psize (float): Pixel size to include in MRC header.
            overwrite (bool, optional): If True, overwrite existing files.
                Defaults to False.
        """
        target_path = PurePosixPath(self.uid) / target_path
        return self.cs.upload_mrc(self.project_uid, target_path, data, psize, overwrite=overwrite)

    def mkdir(
        self,
        target_path: Union[str, PurePosixPath],
        parents: bool = False,
        exist_ok: bool = False,
    ):
        """
        Create a folder in the given job.

        Args:
            target_path (str | Path): Name or path of folder to create inside
                the job directory.
            parents (bool, optional): If True, any missing parents are created
                as needed. Defaults to False.
            exist_ok (bool, optional): If True, does not raise an error for
                existing directories. Still raises if the target path is not a
                directory. Defaults to False.
        """
        self.cs.mkdir(
            project_uid=self.project_uid,
            target_path=PurePosixPath(self.uid) / target_path,
            parents=parents,
            exist_ok=exist_ok,
        )

    def cp(self, source_path: Union[str, PurePosixPath], target_path: Union[str, PurePosixPath] = ""):
        """
        Copy a file or folder into the job directory.

        Args:
            source_path (str | Path): Relative or absolute path of source file
                or folder to copy. If relative, assumed to be within the job
                directory.
            target_path (str | Path, optional): Name or path in the job
                directory to copy into. If not specified, uses the same file
                name as the source. Defaults to "".
        """
        self.cs.cp(
            project_uid=self.project_uid,
            source_path=PurePosixPath(self.uid) / source_path,
            target_path=PurePosixPath(self.uid) / target_path,
        )

    def symlink(self, source_path: Union[str, PurePosixPath], target_path: Union[str, PurePosixPath] = ""):
        """
        Create a symbolic link in job's directory.

        Args:
            source_path (str | Path): Relative or absolute path of source file
                or folder to create a link to. If relative, assumed to be within
                the job directory.
            target_path (str | Path): Name or path of new symlink in the job
                directory. If not specified, creates link with the same file
                name as the source. Defaults to "".
        """
        self.cs.symlink(
            project_uid=self.project_uid,
            source_path=PurePosixPath(self.uid) / source_path,
            target_path=PurePosixPath(self.uid) / target_path,
        )

    def subprocess(
        self,
        args: Union[str, list],
        mute: bool = False,
        checkpoint: bool = False,
        checkpoint_line_pattern: Union[str, Pattern[str], None] = None,
        **kwargs,
    ):
        """
        Launch a subprocess and write its text-based output and error to the job
        log.

        Args:
            args (str | list): Process arguments to run
            mute (bool, optional): If True, does not also forward process output
                to standard output. Defaults to False.
            checkpoint (bool, optional): If True, creates a checkpoint in the
                job event log just before process output begins. Defaults to
                False.
            checkpoint_line_pattern (str | Pattern[str], optional): Regular
                expression to match checkpoint lines for processes with a lot of
                output. If a process outputs a line that matches this pattern, a
                checkpoint is created in the event log before this line is
                forwarded. Defaults to None.
            **kwargs: Additional keyword arguments for ``subprocess.Popen``.

        Raises:
            TypeError: For invalid arguments
            RuntimeError: If process exists with non-zero status code
        """
        import re
        import subprocess

        pttrn = None
        if checkpoint_line_pattern and isinstance(checkpoint_line_pattern, str):
            pttrn = re.compile(checkpoint_line_pattern)
        elif isinstance(checkpoint_line_pattern, re.Pattern):
            pttrn = checkpoint_line_pattern
        elif checkpoint_line_pattern:
            raise TypeError(f"Invalid checkpoint_line_pattern argument type: {type(checkpoint_line_pattern)}")

        args = args if isinstance(args, str) else list(map(str, args))
        with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, **kwargs) as proc:
            assert proc.stdout, f"Subprocess {args} has no standard output"
            if checkpoint:
                self.log_checkpoint()

            self.log("─────── Forwarding subprocess output for the following command ───────")
            self.log(str(args))
            self.log("──────────────────────────────────────────────────────────────────────")

            for line in proc.stdout:
                line = line.rstrip()
                if pttrn and pttrn.match(line):
                    self.log_checkpoint()
                self.log(line)
                if not mute:
                    print(line)

            while proc.poll() is None:
                sleep(1)

            if proc.returncode != 0:
                msg = f"Subprocess {args} exited with status {proc.returncode}"
                self.log(msg, level="error")
                raise RuntimeError(msg)

            self.log("─────────────────────── Subprocess complete. ─────────────────────────")

    def print_param_spec(self):
        """
        Print a table of parameter keys, their title, type and default to
        standard output:

        Examples:

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> job.type
            'extract_micrographs_multi'
            >>> job.print_param_spec()
            Param                       | Title                 | Type    | Default
            =======================================================================
            box_size_pix                | Extraction box size   | integer  | 256
            bin_size_pix                | Fourier crop box size | integer  | None
            compute_num_gpus            | Number of GPUs        | integer  | 1
            ...

        """
        headings = ["Param", "Title", "Type", "Default"]
        rows = []
        for key, details in self.full_spec.params.items():
            if details.hidden is True:
                continue
            type, format = (details.type, details.format)
            if details.anyOf:
                type, format = (details.anyOf[0].type, details.anyOf[0].format)
            display_type = format or type or "any"
            # for an unknown reason, boolean defaults appear as 0/1 instead of True/False
            display_default = repr(bool(details.default)) if display_type == "boolean" else repr(details.default)
            rows.append([key, details.title or key, display_type, display_default])
        print_table(headings, rows)

    def print_input_spec(self):
        """
        Print a table of input keys, their title, type, connection requirements
        and details about their low-level required slots.

        The "Required?" heading also shows the number of outputs that must be
        connected to the input for this job to run.

        Examples:

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> job.type
            'extract_micrographs_multi'
            >>> job.print_output_spec()
            Input       | Title       | Type     | Required? | Input Slots     | Slot Types      | Slot Required?
            =====================================================================================================
            micrographs | Micrographs | exposure | ✓ (1+)    | micrograph_blob | micrograph_blob | ✓
                        |             |          |           | mscope_params   | mscope_params   | ✓
                        |             |          |           | background_blob | stat_blob       | ✕
                        |             |          |           | ctf             | ctf             | ✕
            particles   | Particles   | particle | ✕ (0+)    | location        | location        | ✓
                        |             |          |           | alignments2D    | alignments2D    | ✕
                        |             |          |           | alignments3D    | alignments3D    | ✕
        """
        headings = ["Input", "Title", "Type", "Required?", "Input Slots", "Slot Types", "Slot Required?"]
        rows = []
        for key, input in self.model.spec.inputs.root.items():
            name, title, type = key, input.title, input.type
            required = f"✓ ({input.count_min}" if input.count_min else "✕ (0"
            if input.count_max in (0, "inf"):
                required += "+)"  # unlimited connections
            elif input.count_min == input.count_max:
                required += ")"
            else:
                required += f"-{input.count_max})"
            for slot in input.slots:
                slot = as_input_slot(slot)
                rows.append([name, title, type, required, slot.name, slot.dtype, "✓" if slot.required else "✕"])
                name, title, type, required = ("",) * 4  # only show group info on first iter
        print_table(headings, rows)

    def print_output_spec(self):
        """
        Print a table of output keys, their title, type and details about their
        low-level results.

        Examples:

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> job = cs.find_job("P3", "J42")
            >>> job.type
            'extract_micrographs_multi'
            >>> job.print_output_spec()
            Output                 | Title       | Type     | Result Slots           | Result Types    | Passthrough?
            =========================================================================================================
            micrographs            | Micrographs | exposure | micrograph_blob        | micrograph_blob | ✕
                                   |             |          | micrograph_blob_non_dw | micrograph_blob | ✓
                                   |             |          | background_blob        | stat_blob       | ✓
                                   |             |          | ctf                    | ctf             | ✓
                                   |             |          | ctf_stats              | ctf_stats       | ✓
                                   |             |          | mscope_params          | mscope_params   | ✓
            particles              | Particles   | particle | blob                   | blob            | ✕
                                   |             |          | ctf                    | ctf             | ✕
        """
        headings = ["Output", "Title", "Type", "Result Slots", "Result Types", "Passthrough?"]
        rows = []
        for key, output in self.model.spec.outputs.root.items():
            name, title, type = key, output.title, output.type
            for result in output.results:
                rows.append([name, title, type, result.name, result.dtype, "✓" if result.passthrough else "✕"])
                name, title, type = "", "", ""  # only these print once per group
        print_table(headings, rows)


class ExternalJobController(JobController):
    """
    Mutable custom output job with customizeble input slots and output results.
    Use External jobs to save data save cryo-EM data generated by a software
    package outside of CryoSPARC.

    Created external jobs may be connected to any other CryoSPARC job result as
    an input. Its outputs must be created manually and may be configured to
    passthrough inherited input fields, just as with regular CryoSPARC jobs.

    Create a new External Job with :py:meth:`project.create_external_job() <cryosparc.controllers.project.ProjectController.create_external_job>`.
    or :py:meth:`workspace.create_external_job() <cryosparc.controllers.workspace.WorkspaceController.create_external_job>`.
    ``ExternalJobController`` is a subclass of :py:class:`JobController`
    and inherits all its methods and attributes.

    Examples:

        Import multiple exposure groups into a single job

        >>> from cryosparc.tools import CryoSPARC
        >>> cs = CryoSPARC("http://localhost:61000")
        >>> project = cs.find_project("P3")
        >>> job = project.create_external_job("W3", title="Import Image Sets")
        >>> job.start()
        >>> for i in range(3):
        ...     dset = job.add_output(
        ...         type="exposure",
        ...         name=f"images_{i}",
        ...         slots=["movie_blob", "mscope_params", "gain_ref_blob"],
        ...         alloc=10  # allocate a dataset for this output with 10 rows
        ...     )
        ...     dset['movie_blob/path'] = ...  # populate dataset
        ...     job.save_output(output_name, dset)
        ...
        >>> job.stop()
    """

    def __init__(self, cs: "CryoSPARC", job: Union[Tuple[str, str], Job]) -> None:
        super().__init__(cs, job)
        if self.model.spec.type != "snowflake":
            raise TypeError(f"Job {self.model.project_uid}-{self.model.uid} is not an external job")

    def add_input(
        self,
        type: Datatype,
        name: Optional[str] = None,
        min: int = 0,
        max: Union[int, Literal["inf"]] = "inf",
        slots: Sequence[SlotSpec] = [],
        title: Optional[str] = None,
        desc: Optional[str] = None,
    ):
        """
        Add an input slot to the current job. May be connected to zero or more
        outputs from other jobs (depending on the min and max values).

        Args:
            type (Datatype): cryo-EM data type for this output, e.g., "particle"
            name (str, optional): Output name key, e.g., "picked_particles".
                Same as ``type`` if not specified. Defaults to None.
            min (int, optional): Minimum number of required input connections.
                Defaults to 0.
            max (int | Literal["inf"], optional): Maximum number of input
                connections. Specify ``"inf"`` for unlimited connections.
                Defaults to "inf".
            slots (list[SlotSpec], optional): List of slots that should
                be connected to this input, such as ``"location"`` or  ``"blob"``.
                When connecting the input, if the source job output is missing
                these slots, the external job cannot start or accept outputs.
                Defaults to [].
            title (str, optional): Human-readable title for this input. Defaults
                to name.
            desc (str, optional): Human-readable description for this input.
                Defaults to None.

        Raises:
            CommandError: General CryoSPARC network access error such as
                timeout, URL or HTTP
            InvalidSlotsError: slots argument is invalid

        Returns:
            str: name of created input

        Examples:

            Create an external job that accepts micrographs as input:

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.create_external_job("W1", title="Custom Picker")
            >>> job.uid
            "J3"
            >>> job.add_input(
            ...     type="exposure",
            ...     name="input_micrographs",
            ...     min=1,
            ...     slots=["micrograph_blob", "ctf"],
            ...     title="Input micrographs for picking
            ... )
            "input_micrographs"
        """
        if name and not re.fullmatch(GROUP_NAME_PATTERN, name):
            raise ValueError(
                f'Invalid input name "{name}"; may only contain letters, numbers and underscores, '
                "and must start with a letter"
            )
        if not slots:
            raise ValueError("Must must provide slots=[...] argument with at least one slot")
        if any(isinstance(s, dict) and "prefix" in s for s in slots):
            warnings.warn("'prefix' slot key is deprecated. Use 'name' instead.", DeprecationWarning, stacklevel=2)
        if not name:
            name = type
        if not title:
            title = name
        self.model = self.cs.api.jobs.add_external_input(
            self.project_uid,
            self.uid,
            name,
            InputSpec(
                type=type,
                title=title,
                description=desc or "",
                slots=[as_input_slot(slot) for slot in slots],
                count_min=min,
                count_max=max,
            ),
        )
        return name

    # fmt: off
    @overload
    def add_output(self, type: Datatype, name: Optional[str] = ..., slots: Sequence[SlotSpec] = ..., passthrough: Optional[str] = ..., title: Optional[str] = ...) -> str: ...
    @overload
    def add_output(self, type: Datatype, name: Optional[str] = ..., slots: Sequence[SlotSpec] = ..., passthrough: Optional[str] = ..., title: Optional[str] = ..., *, alloc: Union[int, Dataset]) -> Dataset: ...
    # fmt: on
    def add_output(
        self,
        type: Datatype,
        name: Optional[str] = None,
        slots: Sequence[SlotSpec] = [],
        passthrough: Optional[str] = None,
        title: Optional[str] = None,
        *,
        alloc: Union[int, Dataset, None] = None,
    ) -> Union[str, Dataset]:
        """
        Add an output slot to the current job. Optionally returns the
        corresponding empty dataset if ``alloc`` is specified.

        Args:
            type (Datatype): cryo-EM datatype for this output, e.g., "particle"
            name (str, optional): Output name key, e.g., "selected_particles".
                Same as ``type`` if not specified. Defaults to None.
            slots (list[SlotSpec], optional): List of slot expected to be
                created for this output, such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to [].
            passthrough (str, optional): Indicates that this output inherits
                slots from an existing input with the specified name. The input
                must first be added with ``add_input()``. Defaults to False.
            title (str, optional): Human-readable title for this input. Defaults
                to None.
            alloc (int | Dataset, optional): If specified, pre-allocate and
                return a dataset with the requested slots. Specify an integer
                to allocate a specific number of rows. Specify a Dataset from
                which to inherit unique row IDs (useful when adding passthrough
                outputs). Defaults to None.

        Raises:
            CommandError: General CryoSPARC network access error such as
                timeout, URL or HTTP
            InvalidSlotsError: slots argument is invalid

        Returns:
            str | Dataset: Name of the created output. If ``alloc`` is
                specified as an integer, instead returns blank dataset with the
                given size and random UIDs. If ``alloc`` is specified as a
                Dataset, returns blank dataset with the same UIDs.

        Examples:

            Create and allocate an output for new particle picks

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> particles_dset = job.add_output(
            ...     type="particle",
            ...     name="picked_particles",
            ...     slots=["location", "pick_stats"],
            ...     alloc=10000
            ... )

            Create an inheritied output for input micrographs

            >>> job.add_output(
            ...     type="exposures",
            ...     name="picked_micrographs",
            ...     passthrough="input_micrographs",
            ...     title="Passthrough picked micrographs"
            ... )
            "picked_micrographs"

            Create an output with multiple slots of the same type

            >>> job.add_output(
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"name": "alignments_class_0", "dtype": "alignments3D", "required": True},
            ...         {"name": "alignments_class_1", "dtype": "alignments3D", "required": True},
            ...         {"name": "alignments_class_2", "dtype": "alignments3D", "required": True},
            ...     ]
            ... )
            "particle_alignments"
        """
        if name and not re.fullmatch(GROUP_NAME_PATTERN, name):
            raise ValueError(
                f'Invalid output name "{name}"; may only contain letters, numbers and underscores, '
                "and must start with a letter"
            )
        if not slots:
            raise ValueError("Must must provide slots=[...] argument with at least one slot")
        if any(isinstance(s, dict) and "prefix" in s for s in slots):
            warnings.warn("'prefix' slot key is deprecated. Use 'name' instead.", DeprecationWarning, stacklevel=2)
        if not name:
            name = type
        if not title:
            title = name
        self.model = self.cs.api.jobs.add_external_output(
            self.project_uid,
            self.uid,
            name,
            OutputSpec(type=type, title=title, slots=[as_output_slot(slot) for slot in slots], passthrough=passthrough),
        )
        return name if alloc is None else self.alloc_output(name, alloc)

    def connect(
        self,
        target_input: str,
        source_job_uid: str,
        source_output: str,
        *,
        slots: Sequence[SlotSpec] = [],
        title: Optional[str] = None,
        desc: Optional[str] = None,
        **kwargs,
    ) -> bool:
        """
        Connect the given input for this job to an output with given job UID and
        name. If this input does not exist, it will be added with the given
        slots.

        Args:
            target_input (str): Input name to connect into. Will be created if
                does not already exist.
            source_job_uid (str): Job UID to connect from, e.g., "J42"
            source_output (str): Job output name to connect from , e.g.,
                "particles"
            slots (list[SlotSpec], optional): List of input slots (e.g.,
                "particle" or "blob") to explicitly required for the created
                input. If the given source job is missing these slots, the
                external job cannot start or accept outputs. Defaults to [].
            title (str, optional): Human readable title for created input.
                Defaults to target input name.
            desc (str, optional): Human readable description for created input.
                Defaults to "".

        Raises:
            CommandError: General CryoSPARC network access error such as
                timeout, URL or HTTP
            InvalidSlotsError: slots argument is invalid

        Examples:

            Connect J3 to CTF-corrected micrographs from J2's ``micrographs``
            output.

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> job.connect("input_micrographs", "J2", "micrographs")

        """
        if "refresh" in kwargs:
            warnings.warn("refresh argument no longer applies", DeprecationWarning, stacklevel=2)
        if source_job_uid == self.uid:
            raise ValueError(f"Cannot connect job {self.uid} to itself")
        source_job = self.cs.api.jobs.find_one(self.project_uid, source_job_uid)
        if source_output not in source_job.spec.outputs.root:
            raise ValueError(f"Source job {source_job_uid} does not have output {source_output}")
        output = source_job.spec.outputs.root[source_output]
        if target_input not in self.model.spec.inputs.root:
            if any(isinstance(s, dict) and "prefix" in s for s in slots):
                warnings.warn("'prefix' slot key is deprecated. Use 'name' instead.", DeprecationWarning, stacklevel=2)
                # convert to prevent from warning again
                slots = [as_input_slot(slot) for slot in slots]  # type: ignore
            self.add_input(output.type, target_input, min=1, slots=slots, title=title, desc=desc)
        return super().connect(target_input, source_job_uid, source_output)

    def alloc_output(
        self, name: str, alloc: Union[int, "ArrayLike", Dataset] = 0, *, dtype_params: Dict[str, Any] = {}
    ) -> Dataset:
        """
        Allocate an empty dataset for the given output with the given name.
        Initialize with the given number of empty rows. The result may be
        used with ``save_output`` with the same output name.

        Args:
            name (str): Name of job output to allocate
            size (int | ArrayLike | Dataset, optional): Specify as one of the
                following: (A) integer to allocate a specific number of rows,
                (B) a numpy array of numbers to use for UIDs in the allocated
                dataset or (C) a dataset from which to inherit unique row IDs
                (useful  for allocating passthrough outputs). Defaults to 0.
            dtype_params (dict, optional): Data type parameters when allocating
                results with dynamic column sizes such as ``particle`` ->
                ``alignments3D_multi``. Defaults to {}.

        Returns:
            Dataset: Empty dataset with the given number of rows

        Examples:

            Allocate a dataset of size 10,000 for an output for new particle
            picks

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> job.alloc_output("picked_particles", 10000)
            Dataset([  # 10000 items, 11 fields
                ("uid": [...]),
                ("location/micrograph_path", ["", ...]),
                ...
            ])

            Allocate a dataset from an existing input passthrough dataset

            >>> input_micrographs = job.load_input("input_micrographs")
            >>> job.alloc_output("picked_micrographs", input_micrographs)
            Dataset([  # same "uid" field as input_micrographs
                ("uid": [...]),
            ])

        """
        expected_fields = self.cs.api.jobs.get_output_fields(self.project_uid, self.uid, name, dtype_params)
        if isinstance(alloc, int):
            return Dataset.allocate(alloc, expected_fields)
        elif isinstance(alloc, Dataset):
            return Dataset({"uid": alloc["uid"]}).add_fields(expected_fields)
        else:
            return Dataset({"uid": alloc}).add_fields(expected_fields)

    def save_output(
        self,
        name: str,
        dataset: Dataset,
        *,
        version: int = 0,
        image: Optional[FileOrFigure] = None,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
        **kwargs,
    ):
        """
        Save output dataset to external job.

        Args:
            name (str): Name of output on this job.
            dataset (Dataset): Value of output with only required fields.
            version (int, optional): Version number, when saving multiple
                intermediate iterations. Only the last saved version is kept.
                Defaults to 0.
            image (str | Path | IO | Figure, optional): Optional image file
                or matplotlib Figure to set as the thumbnail for this output.
                Defaults to None.
            savefig_kw (dict, optional): Additional keyword arguments to pass
                to ``figure.savefig()`` when saving matplotlib Figures. Defaults
                to ``dict(bbox_inches="tight", pad_inches=0)``.

        Examples:

            Save a previously-allocated output.

            >>> cs = CryoSPARC("http://localhost:61000")
            >>> project = cs.find_project("P3")
            >>> job = project.find_external_job("J3")
            >>> particles = job.alloc_output("picked_particles", 10000)
            >>> job.save_output("picked_particles", particles)

        """
        if "refresh" in kwargs:
            warnings.warn("refresh argument no longer applies", DeprecationWarning, stacklevel=2)
        self.model = self.cs.api.jobs.save_output(self.project_uid, self.uid, name, dataset, version=version)
        if not image:
            return
        try:
            self.set_output_image(name, image, savefig_kw=savefig_kw)
        except Exception as e:
            warnings.warn(f"Could not upload output thumbnail image due to {e}", stacklevel=2)

    def set_output_image(
        self,
        name: str,
        image: Union[FileOrFigure, GridFSAsset],
        *,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ):
        """
        Set the output image for the given output to the given image file or matplotlib Figure.
        Args:
            name (str): Name of output to set image for.
            image (str | Path | IO | Figure): Image file or matplotlib Figure.

        """
        assets = (
            [image]
            if isinstance(image, GridFSAsset)
            else self._upload_plot(image, name=name, formats=["png"], savefig_kw=savefig_kw)
        )
        self.model = self.cs.api.jobs.set_output_image(self.project_uid, self.uid, name, assets[0])

    def set_tile_image(
        self,
        image: Union[FileOrFigure, GridFSAsset],
        *,
        name: str = "tile0",
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ):
        """
        Set the job tile image to the given image file or matplotlib Figure.

        Args:
            image (str | Path | IO | Figure): Image file or matplotlib Figure.
        """
        assets = (
            [image]
            if isinstance(image, GridFSAsset)
            else self._upload_plot(image, name=name, formats=["png"], savefig_kw=savefig_kw)
        )
        self.model = self.cs.api.jobs.set_tile_image(self.project_uid, self.uid, assets[0])

    def start(self, status: Literal["running", "waiting"] = "waiting"):
        """
        Set job status to "running" or "waiting"

        Args:
            status (str, optional): "running" or "waiting". Defaults to "waiting".
        """
        self.model = self.cs.api.jobs.mark_running(self.project_uid, self.uid, status=status)

    def stop(self, error: str = ""):
        """
        Set job status to "completed" or "failed" if there was an error.

        Args:
            error (str, optional): Error message, will add to event log and set
                job to status to failed if specified. Defaults to "".
        """
        if isinstance(error, bool):  # allowed bool in previous version
            warnings.warn("error should be specified as a string", DeprecationWarning, stacklevel=2)
            error = "An error occurred" if error else ""
        if error:
            self.model = self.cs.api.jobs.mark_failed(self.project_uid, self.uid, error=error)
        else:
            self.model = self.cs.api.jobs.mark_completed(self.project_uid, self.uid)

    @contextmanager
    def run(self):
        """
        Start a job within a context manager and stop the job when the context
        ends.

        Yields:
            ExternalJob: self.

        Examples:

            Job will be marked as "failed" if the contents of the block throw an
            exception

            >>> with job.run():
            ...     job.save_output(...)

        """
        error = ""
        try:
            self.start("running")
            yield self
        except Exception:
            error = traceback.format_exc()
            raise
        finally:
            self.stop(error=error)

    def queue(
        self,
        lane: Optional[str] = None,
        hostname: Optional[str] = None,
        gpus: Sequence[int] = [],
        cluster_vars: Dict[str, Any] = {},
    ):
        raise ExternalJobError(
            "Cannot queue an external job; use `job.start()`/`job.stop()` or `with job.run()` instead"
        )

    def kill(self):
        raise ExternalJobError("Cannot kill an external job; use `job.stop()` instead")



"""Stream processing utilities"""

from abc import ABC, abstractmethod
from pathlib import PurePath
from typing import (
    IO,
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    BinaryIO,
    Iterator,
    List,
    Optional,
    Protocol,
    Union,
    overload,
)

if TYPE_CHECKING:
    from typing_extensions import Buffer, Self

from .constants import EIGHT_MIB
from .util import bopen


class AsyncReadable(Protocol):
    """Any object that has an async read(size) method"""

    def read(self, size: int = ..., /) -> Awaitable[bytes]: ...


class AsyncWritable(Protocol):
    """Any object that has an async write(buffer) method"""

    def write(self, b: "Buffer", /) -> Awaitable[int]: ...


class AsyncBinaryIterator(Protocol):
    """
    Any object that asynchronously yields bytes when iterated e.g.::

        async for chunk in obj:
            print(chunk.decode())
    """

    def __aiter__(self) -> AsyncIterator[bytes]: ...
    def __anext__(self) -> Awaitable[bytes]: ...


class BinaryIteratorIO(BinaryIO, Iterator[bytes]):
    """Read through a iterator that yields bytes as if it was a file"""

    def __init__(self, iter: Iterator[bytes]):
        self._iter = iter
        self._left = b""

    def __iter__(self):
        assert not self._left, "Cannot iterate over a stream that has already been read"
        return iter(self._iter)

    def __next__(self):
        assert not self._left, "Cannot iterate over a stream that has already been read"
        return next(self._iter)

    def readable(self):
        return True

    def seekable(self) -> bool:
        return False

    def writable(self) -> bool:
        return False

    def _read1(self, n: Optional[int] = None):
        while not self._left:
            try:
                self._left = next(self._iter)
            except StopIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    def read(self, n: Optional[int] = None):
        out = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                out.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                out.append(m)
        return b"".join(out)

    def close(self) -> None:
        raise NotImplementedError

    def fileno(self) -> int:
        raise NotImplementedError

    def flush(self) -> None:
        raise NotImplementedError

    def isatty(self) -> bool:
        raise NotImplementedError

    def readline(self, limit: int = -1, /) -> bytes:
        raise NotImplementedError

    def readlines(self, hint: int = -1, /) -> List[bytes]:
        raise NotImplementedError

    def write(self, s, /) -> int:
        raise NotImplementedError

    def writelines(self, lines, /) -> None:
        raise NotImplementedError

    def seek(self, offset: int, whence: int = 0, /) -> int:
        raise NotImplementedError

    def tell(self) -> int:
        raise NotImplementedError

    def truncate(self, size: Optional[int] = None, /) -> int:
        raise NotImplementedError

    def __enter__(self) -> BinaryIO:
        raise NotImplementedError

    def __exit__(self, *args) -> None:
        raise NotImplementedError


class AsyncBinaryIteratorIO(AsyncReadable, AsyncBinaryIterator, AsyncIterator[bytes]):
    """Similar to BinaryIteratorIO except the iterator yields bytes asynchronously"""

    def __init__(self, iter: AsyncBinaryIterator):
        self._iter = iter
        self._left = b""

    def __aiter__(self):
        assert not self._left, "Cannot iterate over a stream that has already been read"
        return self._iter.__aiter__()

    def __anext__(self):
        assert not self._left, "Cannot iterate over a stream that has already been read"
        return self._iter.__anext__()

    def readable(self):
        return True

    def seekable(self) -> bool:
        return False

    async def _read1(self, n: Optional[int] = None):
        while not self._left:
            try:
                self._left = await self._iter.__anext__()
            except StopAsyncIteration:
                break
        ret = self._left[:n]
        self._left = self._left[len(ret) :]
        return ret

    async def read(self, n: Optional[int] = None):
        out = []
        if n is None or n < 0:
            while True:
                m = await self._read1()
                if not m:
                    break
                out.append(m)
        else:
            while n > 0:
                m = await self._read1(n)
                if not m:
                    break
                n -= len(m)
                out.append(m)
        return b"".join(out)


class Streamable(ABC):
    media_type = "application/octet-stream"
    """
    May override in subclasses to derive correct stream type, e.g.,
    "application/x-cryosparc-dataset"
    """

    @classmethod
    def api_schema(cls):
        """
        Schema to use when an API endpoint returns or requests this streamable
        instance in the request or response body.
        """
        return {
            "description": f"A binary stream representing a {cls.__name__} class instance",
            "content": {cls.media_type: {"schema": {"title": cls.__name__, "type": "string", "format": "binary"}}},
        }

    @classmethod
    @abstractmethod
    def load(cls, file: Union[str, PurePath, IO[bytes]], *, media_type: Optional[str] = None) -> "Self":
        """
        Load stream from a file path or readable byte stream. The stream must
        at least implement the `read(size)` function.
        """
        ...

    @classmethod
    def from_iterator(cls, source: Iterator[bytes], *, media_type: Optional[str] = None):
        return cls.load(BinaryIteratorIO(source), media_type=media_type)

    @classmethod
    @abstractmethod
    async def from_async_stream(cls, stream: AsyncReadable, *, media_type: Optional[str] = None) -> "Self":
        """
        Asynchronously load from the given binary stream. The given stream
        parameter must at least have ``async read(n: int | None) -> bytes`` method.
        """
        ...

    @classmethod
    async def from_async_iterator(cls, iterator: AsyncBinaryIterator, *, media_type: Optional[str] = None):
        return await cls.from_async_stream(AsyncBinaryIteratorIO(iterator), media_type=media_type)

    @abstractmethod
    def stream(self) -> Iterator[bytes]: ...

    async def astream(self) -> AsyncIterator[bytes]:
        for chunk in self.stream():
            yield chunk

    def save(self, file: Union[str, PurePath, IO[bytes]]):
        with bopen(file, "wb") as f:
            self.dump(f)

    def dump(self, file: IO[bytes]):
        for chunk in self.stream():
            file.write(chunk)

    def dumps(self) -> bytes:
        return b"".join(self.stream())

    async def adump(self, file: Union[IO[bytes], AsyncWritable]):
        async for chunk in self.astream():
            result = file.write(chunk)
            if isinstance(result, Awaitable):
                await result

    async def adumps(self) -> bytes:
        from io import BytesIO

        data = BytesIO()
        await self.adump(data)
        return data.getvalue()


class Stream(Streamable):
    """
    Generic stream that that leaves handling of the stream data to the caller.
    May accept stream data in any streamable format, though async formats
    must be consumed with async functions.
    """

    @overload
    def __init__(self, *, stream: IO[bytes] = ..., media_type: Optional[str] = ...): ...
    @overload
    def __init__(self, *, iterator: Iterator[bytes] = ..., media_type: Optional[str] = ...): ...
    @overload
    def __init__(self, *, astream: AsyncReadable = ..., media_type: Optional[str] = ...): ...
    @overload
    def __init__(self, *, aiterator: AsyncBinaryIterator = ..., media_type: Optional[str] = ...): ...
    def __init__(
        self,
        *,
        stream: Optional[IO[bytes]] = None,
        iterator: Optional[Iterator[bytes]] = None,
        astream: Optional[AsyncReadable] = None,
        aiterator: Optional[AsyncBinaryIterator] = None,
        media_type: Optional[str] = None,
    ):
        if (stream is not None) + (iterator is not None) + (astream is not None) + (aiterator is not None) != 1:
            raise TypeError("Exactly one of stream, iterator, astream or aiterator must be provided")
        self._stream = stream
        self._iterator = iterator
        self._astream = astream
        self._aiterator = aiterator
        self.media_type = media_type or self.media_type

    @property
    def asynchronous(self):
        return (self._astream is not None) or (self._aiterator is not None)

    @classmethod
    def load(cls, file: Union[str, PurePath, IO[bytes]], *, media_type: Optional[str] = None):
        stream = open(file, "rb") if isinstance(file, (str, PurePath)) else file
        return cls(stream=stream, media_type=media_type)

    @classmethod
    def from_iterator(cls, source: Iterator[bytes], *, media_type: Optional[str] = None):
        return cls(iterator=source, media_type=media_type)

    @classmethod
    async def from_async_stream(cls, stream: AsyncReadable, *, media_type: Optional[str] = None):
        return cls(astream=stream, media_type=media_type)

    @classmethod
    async def from_async_iterator(cls, iterator: AsyncBinaryIterator, *, media_type: Optional[str] = None):
        return cls(aiterator=iterator, media_type=media_type)

    def stream(self) -> Iterator[bytes]:
        if self._stream:
            while chunk := self._stream.read(EIGHT_MIB):
                yield chunk
        elif self._iterator:
            for chunk in self._iterator:
                yield chunk
        else:
            raise TypeError("This is an asynchronous stream, must use astream() instead")

    async def astream(self) -> AsyncIterator[bytes]:
        if self._stream or self._iterator:
            for chunk in self.stream():
                yield chunk
        elif self._astream:
            while chunk := await self._astream.read(EIGHT_MIB):
                yield chunk
        elif self._aiterator:
            async for chunk in self._aiterator:
                yield chunk



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from .session_params import LivePreprocessingParams


class CTF(BaseModel):
    """
    CTF estimation results associated with an exposure.
    """

    accel_kv: float
    """
    """
    amp_contrast: float
    """
    """
    cross_corr_ctffind4: float
    """
    """
    cs_mm: float
    """
    """
    ctf_fit_to_A: float
    """
    """
    df1_A: float
    """
    """
    df2_A: float
    """
    """
    df_angle_rad: float
    """
    """
    exp_group_id: int
    """
    """
    fig_of_merit_gctf: float
    """
    """
    path: str
    """
    """
    phase_shift_rad: float
    """
    """
    type: str
    """
    """


class CtfStats(BaseModel):
    """
    CTF statistics associated with an exposure.
    """

    cross_corr: int
    """
    """
    ctf_fit_to_A: float
    """
    """
    df_range: List[Any]
    """
    """
    df_tilt_normal: List[Any]
    """
    """
    diag_image_path: str
    """
    """
    fit_data_path: str
    """
    """
    ice_thickness_rel: float
    """
    """
    spectrum_dim: int
    """
    """
    type: str
    """
    """


class ECLExposureProperties(BaseModel):
    """
    :meta private:
    """

    do_athena_results_upload: bool = False
    """
    """


class StatBlob(BaseModel):
    """
    Metadata for background blob .mrc file associated with an exposure.
    """

    binfactor: int
    """
    """
    idx: int
    """
    """
    path: str
    """
    """
    psize_A: float
    """
    """
    shape: List[int]
    """
    """


class GainRefBlob(BaseModel):
    """
    Metadata for gain reference blob file associated with an exposure.
    """

    flip_x: int
    """
    """
    flip_y: int
    """
    """
    idx: int
    """
    """
    path: str
    """
    """
    rotate_num: int
    """
    """
    shape: List[int]
    """
    """


class MicrographBlob(BaseModel):
    """
    Metadata for motion-corrected micrograph blob file associated with an exposure.
    """

    format: str
    """
    """
    idx: int
    """
    """
    is_background_subtracted: bool
    """
    """
    path: str
    """
    """
    psize_A: float
    """
    """
    shape: List[int]
    """
    """


class MovieBlob(BaseModel):
    """
    Metadata for raw movie blob file associated with an exposure.
    """

    format: str
    """
    """
    has_defect_file: bool = False
    """
    """
    import_sig: str = "0"
    """
    Unique 64 bit import signature for this exposure, created by computing the
    sha1 of the original absolute path and taking the first 8 bytes.

    Encoded as a string when saved to mongo or JSON to avoid 32 bit overflow
    errors. When using instances of the MovieBlob model from Python,
    ``import_sig`` should be treated as a regular int.
    """
    is_gain_corrected: bool
    """
    """
    path: str
    """
    """
    psize_A: float
    """
    """
    shape: List[int]
    """
    """


class MScopeParams(BaseModel):
    """
    Microscope parameters associated with an exposure.
    """

    accel_kv: float
    """
    """
    beam_shift: List[int] = [0, 0]
    """
    """
    beam_shift_known: int = 0
    """
    """
    cs_mm: float
    """
    """
    defect_path: Optional[str] = None
    """
    """
    exp_group_id: int
    """
    """
    neg_stain: int = 0
    """
    """
    phase_plate: int
    """
    """
    total_dose_e_per_A2: float = 0
    """
    """


class MotionData(BaseModel):
    """
    Motion correction data associated with an exposure.
    """

    frame_end: int
    """
    """
    frame_start: int
    """
    """
    idx: int
    """
    """
    path: str
    """
    """
    psize_A: float
    """
    """
    type: str
    """
    """
    zero_shift_frame: int
    """
    """


class ExposureElement(BaseModel):
    """
    Outputs produced for a specific exposure.
    """

    background_blob: Optional[StatBlob] = None
    """
    """
    ctf: Optional[CTF] = None
    """
    """
    ctf_stats: Optional[CtfStats] = None
    """
    """
    gain_ref_blob: Optional[GainRefBlob] = None
    """
    """
    micrograph_blob: Optional[MicrographBlob] = None
    """
    """
    micrograph_blob_non_dw: Optional[MicrographBlob] = None
    """
    """
    micrograph_blob_thumb: Optional[MicrographBlob] = None
    """
    """
    movie_blob: Optional[MovieBlob] = None
    """
    """
    mscope_params: Optional[MScopeParams] = None
    """
    """
    rigid_motion: Optional[MotionData] = None
    """
    """
    spline_motion: Optional[MotionData] = None
    """
    """
    uid: str = "0"
    """
    """


class PickerLocations(BaseModel):
    """
    Particle picker locations stored as fractional coordinates.
    ``center_x_frac`` and ``center_y_frac`` always have the same length.
    """

    center_x_frac: List[float] = []
    """
    """
    center_y_frac: List[float] = []
    """
    """


class ParticleManual(BaseModel):
    """
    Manual particle picking results.
    """

    count: int = 0
    """
    """
    fields: List[str] = []
    """
    """
    path: str = "."
    """
    """
    location: PickerLocations = PickerLocations()
    """
    """


class ParticleInfo(BaseModel):
    """
    Particle picking results.
    """

    count: int = 0
    """
    """
    fields: List[str] = []
    """
    """
    path: str = "."
    """
    """


class PickerInfoElement(BaseModel):
    """
    Particle picking results for extracted particles.
    """

    count: int = 0
    """
    """
    fields: List[str] = []
    """
    """
    path: str = "."
    """
    """
    output_shape: Optional[int] = None
    """
    """
    picker_type: Optional[Literal["blob", "template", "manual"]] = None
    """
    """


class ExposureGroups(BaseModel):
    """
    Information about Outputs produced by a specific exposure
    """

    exposure: ExposureElement = ExposureElement()
    """
    Exposures have an list of length 1 values for each field in each prefix
    below.
    """
    particle_manual: ParticleManual = ParticleManual()
    """
    Manual picks have a list of values for each field in each prefix
    """
    particle_blob: ParticleInfo = ParticleInfo()
    """
    Blob picks have a list of fields, a count, and a path to a cs file
    """
    particle_template: ParticleInfo = ParticleInfo()
    """
    Template picks have a list of fields, a count, and a path to a cs file
    """
    particle_deep: Dict[str, Any] = {}
    """
    Unused for now
    :meta private:
    """
    particle_extracted: Union[List[PickerInfoElement], ParticleInfo] = ParticleInfo()
    """
    Starts as empty dict, but gets populated as a list of dictionaries. Used for 2-slot extraction.
    """
    particle_manual_extracted: PickerInfoElement = PickerInfoElement()
    """
    Starts as empty dict, but gets populated as a list of dictionaries
    """


class ExposureAttributes(BaseModel):
    """
    Exposure processing metadata.
    """

    found_at: float = 0
    """
    """
    check_at: float = 0
    """
    """
    motion_at: float = 0
    """
    """
    thumbs_at: float = 0
    """
    """
    ctf_at: float = 0
    """
    """
    pick_at: float = 0
    """
    """
    extract_at: float = 0
    """
    """
    manual_extract_at: float = 0
    """
    """
    ready_at: float = 0
    """
    """
    total_motion_dist: float = 0
    """
    """
    max_intra_frame_motion: float = 0
    """
    """
    average_defocus: float = 0
    """
    """
    defocus_range: float = 0
    """
    """
    astigmatism_angle: float = 0
    """
    """
    astigmatism: float = 0
    """
    """
    phase_shift: float = 0
    """
    """
    ctf_fit_to_A: float = 0
    """
    """
    ice_thickness_rel: float = 0
    """
    """
    df_tilt_angle: float = 0
    """
    """
    total_manual_picks: int = 0
    """
    """
    total_blob_picks: int = 0
    """
    """
    blob_pick_score_median: float = 0
    """
    """
    total_template_picks: int = 0
    """
    """
    template_pick_score_median: float = 0
    """
    """
    total_extracted_particles: int = 0
    """
    """
    total_extracted_particles_manual: int = 0
    """
    """
    total_extracted_particles_blob: int = 0
    """
    """
    total_extracted_particles_template: int = 0
    """
    """


class Exposure(BaseModel):
    """
    A single exposure (movie) within a session, along with associated metadata.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    uid: int
    """
    Unique integer identifier for this exposure within the session. Starts at
    1 for the first found exposure and increments by 1 for each new exposure.
    """
    project_uid: str
    """
    """
    session_uid: str
    """
    """
    exp_group_id: int
    """
    Configured exposure group ID.
    """
    abs_file_path: str
    """
    only used when first reading in this file, otherwise use
    ``groups.exposures.movie_blob.path`` relative to the project directory.
    """
    size: int
    """
    """
    discovered_at: datetime.datetime
    """
    """
    picker_type: Optional[Literal["blob", "template", "manual"]] = None
    """
    """
    deleted: bool = False
    """
    """
    parameter_version: Optional[int] = None
    """
    """
    stage: Literal[
        "go_to_found",
        "found",
        "check",
        "motion",
        "ctf",
        "thumbs",
        "pick",
        "extract",
        "extract_manual",
        "ready",
        "restoring",
        "restoring_motion",
        "restoring_thumbs",
        "restoring_ctf",
        "restoring_extract",
        "restoring_extract_manual",
        "compacted",
    ] = "found"
    """
    """
    fail_count: int = 0
    """
    """
    failed: bool = False
    """
    """
    fail_reason: str = ""
    """
    """
    in_progress: bool = False
    """
    """
    manual_reject: bool = False
    """
    """
    threshold_reject: bool = False
    """
    """
    test: bool = False
    """
    """
    worker_juid: Optional[str] = None
    """
    """
    priority: int = 0
    """
    """
    groups: ExposureGroups = ExposureGroups()
    """
    """
    attributes: ExposureAttributes = ExposureAttributes()
    """
    """
    test_parameters: Optional[LivePreprocessingParams] = None
    """
    """
    preview_img_1x: List[str] = []
    """
    """
    preview_img_2x: List[str] = []
    """
    """
    thumb_shape: List[int] = []
    """
    """
    micrograph_shape: List[int] = []
    """
    """
    micrograph_psize: Optional[float] = None
    """
    """
    ecl: ECLExposureProperties = ECLExposureProperties()
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Any, Dict, Union

from pydantic import BaseModel


class When(BaseModel):
    """
    Boolean expression composition utility class that encodes logic for when job
    metadata (e.g., param hidden state) should be True or False based on the
    state of a job.

    Examples:

        Values for ``expr`` for various common cases:

        When param foo is truthy:
            ``"foo"``

        When param foo equals 42:
            ``{"op": "==", "param": "foo", "value": 42}``

        When param bar is in the list [1, 2, 3]:
            ``{"op": "in", "param": "bar", "value": [1, 2, 3]}``

        When input connection "particles" is present:
            ``{"connection": "particles"}``

        When param foo equals 42 AND input connection "particles" is present:
            ``{"op": "and", "left": {"op": "==", "param": "foo", "value": 42}, "right": {"connection": "particles"}}``
    """

    expr: Union[str, Dict[str, Any]]
    """
    Logical expression encoded as a dictionary or a string (param name).
    """



import warnings
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

from typing_extensions import Unpack

from ..dataset import Dataset
from ..dataset.row import R
from ..models.workspace import Workspace
from ..search import JobSearch
from ..spec import Datatype, SlotSpec
from . import Controller, as_output_slot
from .job import ExternalJobController, FileOrFigure, JobController

if TYPE_CHECKING:
    from ..tools import CryoSPARC


class WorkspaceController(Controller[Workspace]):
    """
    Accessor class to a workspace in CryoSPARC with ability create jobs and save
    results. Should be created with`
    :py:meth:`cs.find_workspace() <cryosparc.tools.CryoSPARC.find_workspace>` or
    :py:meth:`project.find_workspace() <cryosparc.controllers.project.ProjectController.find_workspace>`.

    Arguments:
        workspace (tuple[str, str] | Workspace): either _(Project UID, Workspace UID)_
            tuple or Workspace model, e.g. ``("P3", "W4")``

    Attributes:
        model (Workspace): All workspace data from the CryoSPARC database.
            Contents may change over time, use :py:meth:`refresh` to update.
    """

    uid: str
    """
    Workspace unique ID, e.g., "W42"
    """
    project_uid: str
    """
    Project unique ID, e.g., "P3"
    """

    def __init__(self, cs: "CryoSPARC", workspace: Union[Tuple[str, str], Workspace]) -> None:
        self.cs = cs
        if isinstance(workspace, tuple):
            self.project_uid, self.uid = workspace
            self.refresh()
        else:
            self.project_uid = workspace.project_uid
            self.uid = workspace.uid
            self.model = workspace

    @property
    def title(self) -> Optional[str]:
        """Workspace title"""
        return self.model.title

    @property
    def desc(self) -> Optional[str]:
        """Workspace description"""
        return self.model.description

    def refresh(self):
        """
        Reload this workspace from the CryoSPARC database.

        Returns:
            WorkspaceController: self
        """
        self.model = self.cs.api.workspaces.find_one(self.project_uid, self.uid)
        return self

    def set_title(self, title: str):
        """
        Set the workspace title.

        Args:
            title (str): New workspace title
        """
        self.model = self.cs.api.workspaces.set_title(self.project_uid, self.uid, title=title)

    def set_description(self, desc: str):
        """
        Set the workspace description. May include Markdown formatting.

        Args:
            desc (str): New workspace description
        """
        self.model = self.cs.api.workspaces.set_description(self.project_uid, self.uid, description=desc)

    def find_jobs(self, *, order: Literal[1, -1] = 1, **search: Unpack[JobSearch]) -> Iterable[JobController]:
        """
        Search jobs in the current workspace.

        Example:
            >>> jobs = workspace.find_jobs()  # all jobs in workspace
            >>> jobs = workspace.find_jobs(
            ...     type="homo_reconstruct",
            ...     completed_at=(datetime(2025, 3, 1), datetime(2025, 3, 31)),
            ...     order=-1,
            ... )
            >>> for job in jobs:
            ...     print(job.uid)

        Args:
            **search (JobSearch): Additional search parameters to filter jobs,
                specified as keyword arguments.

        Returns:
            Iterable[JobController]: job accessor objects
        """
        return self.cs.find_jobs(self.project_uid, workspace_uid=self.uid, order=order, **search)

    def find_job(self, job_uid: str) -> JobController:
        """
        Find a job in the current workspace by its UID.

        Args:
            job_uid (str): Job UID to find, e.g., "J42"

        Returns:
            JobController: job accessor object
        """
        jobs = self.cs.api.jobs.find(project_uid=[self.project_uid], workspace_uid=[self.uid], uid=[job_uid], limit=1)
        if len(jobs) == 0:
            raise ValueError(f"Job {job_uid} not found in workspace {self.project_uid}-{self.uid}")
        return JobController(self.cs, jobs[0])

    def create_job(
        self,
        type: str,
        connections: Dict[str, Union[Tuple[str, str], List[Tuple[str, str]]]] = {},
        params: Dict[str, Any] = {},
        title: str = "",
        desc: str = "",
    ) -> JobController:
        """
        Create a new job with the given type. Use
        :py:attr:`cs.job_register <cryosparc.tools.CryoSPARC.job_register>`
        to find available job types on the connected CryoSPARC instance.

        Args:
            project_uid (str): Project UID to create job in, e.g., "P3"
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            type (str): Job type identifier, e.g., "homo_abinit"
            connections (dict[str, tuple[str, str] | list[tuple[str, str]]]):
                Initial input connections. Each key is an input name and each
                value is a (job uid, output name) tuple. Defaults to {}
            params (dict[str, any], optional): Specify parameter values.
                Defaults to {}.
            title (str, optional): Job title. Defaults to "".
            desc (str, optional): Job markdown description. Defaults to "".

        Returns:
            JobController: created job instance. Raises error if job cannot be created.

        Examples:

            Create an Import Movies job.

            >>> from cryosparc.tools import CryoSPARC
            >>> cs = CryoSPARC("http://localhost:61000")
            >>> workspace = cs.find_workspace("P3", "W3")
            >>> import_job = workspace.create_job("W1", "import_movies")
            >>> import_job.set_param("blob_paths", "/bulk/data/t20s/*.tif")
            True

            Create a 3-class ab-initio job connected to existing particles.

            >>> abinit_job = workspace.create_job("homo_abinit"
            ...     connections={"particles": ("J20", "particles_selected")}
            ...     params={"abinit_K": 3}
            ... )
        """
        return self.cs.create_job(
            self.project_uid, self.uid, type, connections=connections, params=params, title=title, desc=desc
        )

    def create_external_job(
        self,
        title: str = "",
        desc: str = "",
    ) -> ExternalJobController:
        """
        Add a new External job to this workspace to save generated outputs to.

        Args:
            workspace_uid (str): Workspace UID to create job in, e.g., "W1"
            title (str, optional): Title for external job (recommended).
                Defaults to "".
            desc (str, optional): Markdown description for external job.
                Defaults to "".

        Returns:
            ExternalJobController: created external job instance
        """
        return self.cs.create_external_job(self.project_uid, self.uid, title, desc)

    def save_external_result(
        self,
        dataset: Dataset[R],
        type: Datatype,
        name: Optional[str] = None,
        slots: Optional[List[SlotSpec]] = None,
        passthrough: Optional[Tuple[str, str]] = None,
        title: str = "",
        desc: str = "",
        image: Optional[FileOrFigure] = None,
        savefig_kw: dict = dict(bbox_inches="tight", pad_inches=0),
    ) -> str:
        """
        Save the given result dataset to a workspace.

        Args:
            dataset (Dataset): Result dataset.
            type (Datatype): Type of output dataset.
            name (str, optional): Name of output on created External job. Same
                as type if unspecified. Defaults to None.
            slots (list[SlotSpec], optional): List of slots expected to
                be created for this output such as ``location`` or ``blob``. Do
                not specify any slots that were passed through from an input
                unless those slots are modified in the output. Defaults to None.
            passthrough (tuple[str, str], optional): Indicates that this output
                inherits slots from the specified output. e.g., ``("J1",
                "particles")``. Defaults to None.
            title (str, optional): Human-readable title for this output.
                Defaults to "".
            desc (str, optional): Markdown description for this output. Defaults
                to "".
            image (str | Path | IO | Figure, optional): Optional image file
                or matplotlib Figure to set as the image for this output.
                Defaults to None.
            savefig_kw (dict, optional): Additional keyword arguments to pass
                to ``figure.savefig()`` when saving matplotlib Figures. Defaults
                to ``dict(bbox_inches="tight", pad_inches=0)``.

        Returns:
            str: UID of created job where this output was saved.

        Examples:

            Save all particle data

            >>> particles = Dataset()
            >>> workspace.save_external_result(particles, 'particle')
            "J43"

            Save new particle locations that inherit passthrough slots from a
            parent job

            >>> particles = Dataset()
            >>> workspace.save_external_result(
            ...     dataset=particles,
            ...     type='particle',
            ...     name='particles',
            ...     slots=['location'],
            ...     passthrough=('J42', 'selected_particles'),
            ...     title='Re-centered particles'
            ... )
            "J44"

            Save a result with multiple slots of the same type.

            >>> workspace.save_external_result(
            ...     dataset=particles,
            ...     type="particle",
            ...     name="particle_alignments",
            ...     slots=[
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_0", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_1", "required": True},
            ...         {"dtype": "alignments3D", "prefix": "alignments_class_2", "required": True},
            ...     ]
            ... )
            "J45"
        """
        if slots and any(isinstance(s, dict) and "prefix" in s for s in slots):
            warnings.warn("'prefix' slot key is deprecated. Use 'name' instead.", DeprecationWarning, stacklevel=2)
            # convert to prevent from warning again
            slots = [as_output_slot(slot) for slot in slots]  # type: ignore
        return self.cs.save_external_result(
            self.project_uid,
            self.uid,
            dataset=dataset,
            type=type,
            name=name,
            slots=slots,
            passthrough=passthrough,
            title=title,
            desc=desc,
            image=image,
            savefig_kw=savefig_kw,
        )



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from .instance import InstanceInformation
from .scheduler_target import Cluster, Node


class SchedulerTargetInformation(BaseModel):
    """
    Information for a configured job scheduler target node, used in diagnostics.
    """

    cache_path: Optional[str] = None
    """
    Path the SSD cache scratch directory, if applicable.
    """
    cache_reserve_mb: Optional[int] = None
    """
    Ensure at least this much space is free on the SSD scratch drive before
    caching.
    """
    cache_quota_mb: Optional[int] = None
    """
    Do not cache more than this amount on the SSD scrath drive..
    """
    lane: str
    """
    Lane name this target belongs to.
    """
    name: str
    """
    Identifier for this target.
    """
    title: str
    """
    Human-readable title for this target.
    """
    desc: Optional[str] = None
    """
    Human-readable description for this target.
    """
    hostname: str
    """
    Network machine hostname (same as name for for clusters).
    """
    worker_bin_path: str
    """
    Path to cryosparc_worker/bin/cryosparcw executable.
    """
    config: Union[Node, Cluster]
    """
    Target configuration details.
    """
    instance_information: Optional[InstanceInformation] = None
    """
    System information for the target node, if available.
    """


class RuntimeDiagnostics(BaseModel):
    """
    Runtime diagnostics information for CryoSPARC instance.
    """

    cryosparc_version: str
    """
    Running version of CryoSPARC, e.g., 'v5.0.0'.
    """
    cryosparc_patch: str
    """
    Applied patch, if any. Set to '' if no patch is applied.
    """
    instance_information: InstanceInformation
    """
    Information about the CryoSPARC instance.
    """
    scheduler_targets: List[SchedulerTargetInformation]
    """
    List of configured job scheduler target nodes.
    """
    db_stats: Dict[str, Any]
    """
    Database statistics snapshot.
    """
    date_generated: datetime.datetime
    """
    UTC datetime when the diagnostics were generated.
    """



import json
import re
import urllib.parse
import warnings
from contextlib import contextmanager
from enum import Enum
from typing import Any, Callable, ClassVar, Dict, Iterator, List, Optional, Tuple, TypedDict, Union

import httpx
from pydantic import BaseModel

from . import registry
from .errors import APIError
from .json_util import api_default, api_encode, api_object_hook
from .models.auth import Token
from .stream import Streamable

_BASE_RESPONSE_TYPES = {"string", "integer", "number", "boolean"}

Auth = Union[str, Tuple[str, str]]
"""
Auth token or email/password.
"""


class APIRequest(TypedDict):
    params: Dict[str, Any]
    headers: Dict[str, str]
    content: Any
    data: Optional[Dict[str, Any]]  # form data
    files: Optional[Dict[str, Any]]


class APINamespace:
    """
    Collection of API methods that call a certain namespace, e.g., only the
    methods under http://master:39004/pipelines/
    """

    _json_encoder: ClassVar[Callable] = api_encode
    _json_decoder: ClassVar[Callable] = api_default

    _client: httpx.Client

    def __init__(self, http_client: httpx.Client):
        self._client = http_client

    def _set_headers(self, update: Dict[str, str]):
        """For testing only, reset client headers"""
        self._client.headers.update(update)

    def _prepare_request_stream(self, streamable):
        return streamable.stream()

    def _construct_request(self, _path: str, _schema, *args, **kwargs) -> Tuple[str, APIRequest]:
        args = list(args)
        query_params = {}
        func_name = _schema["summary"]
        headers = {}
        client_headers = {h for h in self._client.headers.keys()}
        request_body = None
        data = None
        files = None

        for param_schema in sort_params_schema(_path, _schema.get("parameters", [])):
            # Compile function params
            param_name: str = param_schema["name"]
            param_in: str = param_schema["in"]
            assert param_in in (
                "path",
                "query",
                "header",
            ), f"[API] Param specification for '{param_name}' in {param_in} NOT supported."

            if param_in == "path" and args:
                # path param must be specified positionally
                param, args = args[0], args[1:]
                _path = _path.replace("{%s}" % param_name, _uriencode(param))
            elif param_in == "query" and param_name in kwargs and (value := kwargs.pop(param_name)) is not None:
                # query param must be in kwargs
                query_params[param_name] = APINamespace._json_encoder(value)
            elif (
                param_in == "header"
                and (header_name := param_name.replace("-", "_")) in kwargs
                and (value := kwargs.pop(header_name)) is not None
            ):
                # header must be in kwargs
                headers[param_name] = APINamespace._json_encoder(value)
            elif param_in == "header" and param_name in client_headers:
                pass  # in default headers, no action required
            elif param_schema["required"]:
                raise TypeError(f"[API] {func_name}() missing required argument: '{param_name}'")

        if "requestBody" in _schema:
            content_schema = _schema["requestBody"].get("content", {})
            if stream_mime_type := registry.first_streamable_mime(content_schema.keys()):
                body_name = _get_schema_param_name(content_schema[stream_mime_type]["schema"], "body")
                headers["Content-Type"] = stream_mime_type
                if args:
                    streamable, args = args[0], args[1:]
                elif body_name in kwargs:
                    streamable = kwargs.pop(body_name)
                elif content_schema.get("required", False):
                    raise TypeError(f"[API] {func_name}() missing required argument: {body_name}")
                else:
                    streamable = None

                if streamable is not None:
                    if not isinstance(streamable, Streamable):
                        raise TypeError(f"[API] {func_name}() invalid argument {streamable}; expected Streamable type")
                    request_body = self._prepare_request_stream(streamable)
            elif "application/json" in content_schema:
                body_name = _get_schema_param_name(content_schema["application/json"]["schema"], "body")
                headers["Content-Type"] = "application/json"
                if args:
                    request_body, args = args[0], args[1:]
                    request_body = json.dumps(request_body, default=APINamespace._json_decoder)
                elif body_name in kwargs:
                    request_body = kwargs.pop(body_name)
                    request_body = json.dumps(request_body, default=APINamespace._json_decoder)
                elif content_schema.get("required", False):
                    raise TypeError(f"[API] {func_name}() missing required argument: {body_name}")
            elif "application/x-www-form-urlencoded" in content_schema:
                assert kwargs, (
                    f"[API] {func_name}() requires x-www-form-urlencoded which "
                    "does not yet support positional arguments for the content body."
                )
                data, kwargs = kwargs, {}
            elif "multipart/form-data" in content_schema:
                assert kwargs, (
                    f"[API] {func_name}() requires multipart/form-data which "
                    "does not yet support positional arguments for the content body."
                )
                files, kwargs = kwargs, {}
            else:
                raise TypeError(f"[API] Does not yet support request body with content schema {content_schema}")

        # Fail if any extra parameters
        if args:
            raise TypeError(f"[API] {func_name}() given extra positional arguments ({', '.join(map(str, args))})")
        if kwargs:
            raise TypeError(f"[API] {func_name}() given unknown keyword arguments ({', '.join(kwargs.keys())})")

        return _path, APIRequest(params=query_params, headers=headers, content=request_body, data=data, files=files)

    def _handle_response(self, schema, res: httpx.Response):
        responses_schema = schema.get("responses", {})
        response_schema = responses_schema.get(str(res.status_code))
        if not response_schema:
            res.raise_for_status()
            raise APIError("Received unknown response", res=res)

        if "content" not in response_schema:
            res.raise_for_status()
            return None

        content_schema = response_schema["content"]
        stream_mime_type = registry.first_streamable_mime(content_schema.keys())
        if stream_mime_type is not None:  # This is a streaming type
            stream_class = registry.get_stream_class(stream_mime_type)
            assert stream_class
            return stream_class.from_iterator(
                res.iter_bytes(),
                media_type=res.headers.get("Content-Type", stream_mime_type),
            )
        elif "text/plain" in content_schema:
            return res.text
        elif "application/json" in content_schema:
            data = res.json(object_hook=api_object_hook)
            if res.status_code >= 400:
                raise APIError("Received error response", res=res, data=data)
            return _decode_json_response(data, content_schema["application/json"]["schema"])
        else:
            raise APIError(f"Received unimplemented response type in {content_schema.keys()}", res=res)

    @contextmanager
    def _request(self, method: str, url: str, *args, **kwargs) -> Iterator[httpx.Response]:
        attempt = 0
        max_attempts = 3
        while attempt < max_attempts:
            attempt += 1
            try:
                request = self._client.request(method, url, *args, **kwargs)
                yield request
                return
            except httpx.TransportError as err:
                warnings.warn(
                    f"*** API client {method.upper()} {url} failed due to {err} (attempt {attempt}/{max_attempts})"
                )
                if attempt == max_attempts:
                    raise

    @contextmanager
    def _request_stream(self, method: str, url: str, *args, **kwargs) -> Iterator[httpx.Response]:
        attempt = 0
        max_attempts = 3
        while attempt < max_attempts:
            attempt += 1
            try:
                with self._client.stream(method, url, *args, **kwargs) as response:
                    yield response
                    return
            except httpx.TransportError as err:
                warnings.warn(
                    f"*** API client {method.upper()} {url} stream failed due to {err} (attempt {attempt}/{max_attempts})"
                )
                if attempt == max_attempts:
                    raise

        raise RuntimeError(f"[API] Could not complete request {method.upper()} {url}")

    def _call(self, _method: str, _path: str, _schema, *args, **kwargs):
        """Meta-call method that runs whenever a named function is called on
        this namespace"""
        responses_schema = _schema.get("responses", {})
        _path, req = self._construct_request(_path, _schema, *args, **kwargs)

        # Submit a response and get a generator for it. Resulting res_gen should
        # only yield a single item. Cleaned up at the end of this function call.
        content_schema = responses_schema.get("201", {}).get("content", {})
        stream_mime_type = registry.first_streamable_mime(content_schema.keys())

        try:
            ctx = (
                self._request(_method, _path, **req)
                if stream_mime_type is None
                else self._request_stream(_method, _path, **req)
            )
            with ctx as res:
                return self._handle_response(_schema, res)
        except httpx.HTTPStatusError as err:
            raise APIError("received error response", res=err.response) from err


class APIClient(APINamespace):
    """Root API namespace interface for an OpenAPI server"""

    _namespace_class = APINamespace

    def __init__(
        self,
        base_url: Optional[str] = None,
        *,
        auth: Optional[Auth] = None,  # token or email/password
        headers: Optional[Dict[str, str]] = None,
        timeout: float = 300,
        http_client: Optional[httpx.Client] = None,
    ):
        if base_url and http_client:
            raise TypeError(f"Cannot specify both base_url ({base_url}) and http_client ({http_client})")
        if not http_client:
            if not base_url:
                raise TypeError(f"Must specify either base_url ({base_url}) or http_client ({http_client})")
            http_client = httpx.Client(base_url=base_url, timeout=timeout)
        http_client.headers.update(headers)
        super().__init__(http_client)
        self._attrs = set()
        self(auth=auth)  # query the OpenAPI server and populate endpoint functions

    def __del__(self):
        # Clean up client when it gets garbage collected
        self._client.close()

    def __call__(self, auth: Optional[Auth] = None):
        """
        Re-generate all routes and internal namespaces. Optionally accepts
        either an authentication bearer token or an email/password tuple. Note
        that the password should be encoded in sha256.
        """
        self._reset()
        try:
            with self._request("get", "/openapi.json") as res:
                res.raise_for_status()
                schema = res.json()
        except json.JSONDecodeError as e:
            raise ValueError("Error reading JSON response") from e
        self._process_schema(schema)
        if auth:
            self._authorize(auth)
        return schema

    def _reset(self):
        for attr in self._attrs:
            delattr(self, attr)
        self._attrs.clear()

    def _process_schema(self, schema):
        assert isinstance(schema, dict), "[API] Invalid OpenAPI schema response: Not a dictionary"
        for key in {"info", "paths", "components"}:
            assert key in schema, (
                f"[API] Invalid OpenAPI schema response: Missing '{key}' key, got keys {list(schema.keys())}"
            )

        for path, path_schema in schema["paths"].items():
            for method, endpoint_schema in path_schema.items():
                self._register_endpoint(method, path, endpoint_schema)

    def _register_endpoint(self, method: str, path: str, schema):
        """
        CryoSPARC's OpenAPI server is configured such that the "summary"
        property in the provided endpoint schemas at /openapi.json is the name
        of the function in python. '.' used to delimit namespaced endpoints.
        """
        namespace = self
        func_name: str = schema["summary"]
        if "." in func_name:
            namespace_name, func_name = func_name.split(".", 1)
            namespace = self._get_namespace(namespace_name)
        else:
            self._attrs.add(func_name)

        setattr(namespace, func_name, self._generate_endpoint(func_name, namespace, method, path, schema))

    def _generate_endpoint(self, func_name: str, namespace, method: str, path: str, schema):
        def endpoint(*args, **kwargs):
            return namespace._call(method, path, schema, *args, **kwargs)

        endpoint.__name__ = func_name
        return endpoint

    def _get_namespace(self, name: str):
        if not hasattr(self, name):
            setattr(self, name, self._namespace_class(self._client))
            self._attrs.add(name)
        namespace = getattr(self, name)
        assert isinstance(namespace, self._namespace_class), (
            f"{self} name conflict with namespace '{name}'. This is likely a bug"
        )
        return namespace

    def _authorize(self, auth: Auth):
        token = (
            Token(access_token=auth, token_type="bearer")
            if isinstance(auth, str)
            else self.login(grant_type="password", username=auth[0], password=auth[1])  # type: ignore
        )
        self._client.headers["Authorization"] = f"{token.token_type.title()} {token.access_token}"


def sort_params_schema(path: str, param_schema: List[dict]):
    """
    Sort the OpenAPI endpoint parameters schema in order that path params appear
    in the given URI.
    """
    path_params = {p["name"]: p for p in param_schema if p["in"] == "path"}
    known_path_params = re.findall(r"{([^}]*)}", path)
    return [path_params[name] for name in known_path_params] + [p for p in param_schema if p["in"] != "path"]


def _get_schema_param_name(schema: dict, default: str = "param") -> str:
    """
    Given a parameter schema, convert its title to a valid python argument
    identifier. Used to determine kwarg name of body arguments.
    """
    return schema.get("title", default).lower().replace(" ", "_")


def _matches_schema_type(value: Any, schema: dict) -> bool:
    if "$ref" in schema:
        model_class = registry.model_for_ref(schema["$ref"])
        if model_class:
            return True
    schema_type = schema.get("type")
    if schema_type == "object":
        return isinstance(value, dict)
    if schema_type == "array":
        return isinstance(value, list)
    if schema_type in _BASE_RESPONSE_TYPES:
        return isinstance(value, (str, float, int, bool))
    return False


def _decode_json_response(value: Any, schema: dict):
    # Check for empty schema/value, means just return JSON-decoded value as-is
    if value is None or not schema:
        return value

    # Don't attempt parsing unions without a model
    if "anyOf" in schema:
        for subschema in schema["anyOf"]:
            if _matches_schema_type(value, subschema):
                try:
                    return _decode_json_response(value, subschema)
                except (TypeError, ValueError):
                    continue
        warnings.warn("[API] Warning: No union schemas matched response. Returning API result as plain object.")
        return value

    # Check for schema that links to one of our existing models
    if "$ref" in schema:
        model_class = registry.model_for_ref(schema["$ref"])
        if model_class and registry.is_literal(model_class):
            return value  # literal type, return as-is
        elif model_class and issubclass(model_class, Enum):
            return model_class(value)
        elif model_class and issubclass(model_class, dict):  # typed dict
            return model_class(**value)
        elif model_class and issubclass(model_class, BaseModel):  # pydantic model
            # use model_validate in case validator result derives from subtype, e.g., Event model
            return model_class.model_validate(value)
        warnings.warn(
            f"[API] Warning: Received API response with unregistered schema type {schema['$ref']}. "
            "Returning as plain object."
        )
        return value

    # Check for Base case types
    if "type" in schema and schema["type"] in _BASE_RESPONSE_TYPES:
        return value

    # Recursively decode list or tuple
    if "type" in schema and schema["type"] == "array":
        collection_type, items_key = (tuple, "prefixItems") if "prefixItems" in schema else (list, "items")
        return collection_type(_decode_json_response(item, schema[items_key]) for item in value)

    # Recursively decode object
    if "type" in schema and schema["type"] == "object":
        prop_schemas = schema.get("properties", {})
        default_prop_schema = schema.get("additionalProperties", {})
        return {
            key: _decode_json_response(val, prop_schemas.get(key, default_prop_schema)) for key, val in value.items()
        }

    # No other result found, return as plain JSON
    return value


def _uriencode(val: Any):
    # Encode any string-compatible value so that it may be used in a URI.
    return urllib.parse.quote(val if isinstance(val, (str, bytes)) else str(val))



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from .job_spec import BuilderTag, Category, InputSpecs, JobRegisterError, OutputSpecs, Stability
from .params import ParamSection
from .when import When


class JobRegisterParamTypeSpec(BaseModel):
    type: Literal["string", "integer", "number", "boolean", "null", "array"]
    """
    Basic type of the parameter.
    """
    format: Optional[Literal["path", "resources"]] = None
    """
    Expected format or subtype of the parameter, if applicable.
    """


class JobRegisterParamSpec(BaseModel):
    """
    Full specification for a job parameter.
    """

    type: Optional[Literal["string", "integer", "number", "boolean", "null", "array"]] = None
    """
    Basic type of the parameter. Will be empty if 'anyOf' is used.
    """
    anyOf: List[JobRegisterParamTypeSpec] = []
    """
    List of possible types for union or nullable parameters.
    """
    items: Optional[JobRegisterParamTypeSpec] = None
    """
    Specification for items in an 'array' type parameter.
    """
    format: Optional[Literal["path", "resources"]] = None
    """
    Expected format or subtype of the parameter, if applicable.
    """
    default: Union[str, int, float, str, List[str], None] = None
    """
    Default value for the parameter.
    """
    title: Optional[str]
    """
    Human-readable title of the parameter.
    """
    description: Optional[str] = None
    """
    Detailed description of the parameter.
    """
    examples: Optional[List[Any]] = None
    """
    Example values for the parameter.
    """
    gt: Optional[float] = None
    """
    Greater than constraint for numeric parameters.
    """
    ge: Optional[float] = None
    """
    Greater than or equal to constraint for numeric parameters.
    """
    lt: Optional[float] = None
    """
    Less than constraint for numeric parameters.
    """
    le: Optional[float] = None
    """
    Less than or equal to constraint for numeric parameters.
    """
    multiple_of: Optional[float] = None
    """
    Multiple of constraint for numeric parameters.
    """
    enum: Optional[List[Union[str, int, float, str, List[str]]]] = None
    """
    List of allowed values for the parameter, if applicable.
    """
    labels: Optional[List[str]] = None
    """
    Human-readable labels for each value in enum.
    """
    pattern: Optional[str] = None
    """
    Regex pattern that string parameters must match.
    """
    hidden: Union[bool, When] = False
    """
    Whether the parameter is hidden in the UI.
    """
    advanced: bool = False
    """
    Whether the parameter is considered advanced.
    """
    section: Optional[ParamSection] = None
    """
    Section the parameter belongs to in the UI.
    """
    allowed: List[Literal["dir", "file", "glob"]] = []
    """
    Allowed path types for path parameters that have type 'string' and format 'path'.
    """
    validate_path: bool = True
    """
    Whether to validate that paths parameter values exist and are the right allowed type before launching a job.
    """
    legacy_name: Optional[str] = None
    """
    Legacy name of the parameter in previous versions of CryoSPARC, if applicable.
    """
    required_param: bool = False
    """
    Whether the parameter must be set to launch the job.
    """
    ecl_visible: bool = False
    """
    :meta private:
    """


class JobRegisterJobSpec(BaseModel):
    """
    Full specification for a job type in the job register.
    """

    type: str
    """
    Unique identifier for the job type, e.g., 'import_movies'.
    """
    title: str
    """
    Human-readable title of the job type.
    """
    shorttitle: str
    """
    Short title of the job type.
    """
    description: str
    """
    Detailed description of the job type.
    """
    stability: Stability
    """
    Stability level of the job type. 'obsolete' jobs cannot be created or run.
    """
    category: Category
    """
    Category of the job type for grouping in the UI.
    """
    tags: List[BuilderTag] = []
    """
    Job Builder tags associated with the job type.
    """
    hidden: bool = False
    """
    Whether the job type is hidden in the UI.
    """
    interactive: bool = False
    """
    Whether this is an interactive job. Jobs of this type can only run directly on master.
    """
    lightweight: bool = False
    """
    Whether this is a lightweight job. Lightweight jobs never require GPUs
    and can run on either master or workers.
    """
    inputs: InputSpecs
    """
    Input specifications for the job type.
    """
    outputs: OutputSpecs
    """
    Output specifications for the job type.
    """
    params: Dict[str, JobRegisterParamSpec]
    """
    Parameter specifications for the job type.
    """


class JobRegister(BaseModel):
    """
    A registry of all available job types and their specifications.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    specs: List[JobRegisterJobSpec] = []
    """
    List of job type specifications in the registry.
    """
    error: Optional[JobRegisterError] = None
    """
    Error information if there was an issue generating the registry.
    """
    username: Optional[str] = None
    """
    :meta private:
    """
    categories: Dict[str, str]
    """
    Category values to titles map which also indicates order in which jobs
    should be grouped.
    """



from datetime import datetime
from typing import List, Tuple, TypeVar, Union

from typing_extensions import TypedDict

from .models.job import JobStatus
from .models.job_spec import Category

T = TypeVar("T")
Eq = T
In = Union[T, List[T]]
Range = Tuple[T, T]


class JobSearch(TypedDict, total=False):
    """
    Type of keyword arguments for ``find_jobs`` methods.
    """

    type: In[str]
    """
    Include jobs with the given type.
    """
    status: In[JobStatus]
    """
    Include jobs with the given status.
    """
    category: In[Category]
    """
    Include jobs with the given category.
    """
    created_at: Range[datetime]
    """
    Include jobs created within the given time range.
    """
    updated_at: Range[datetime]
    """
    Include jobs updated within the given time range.
    """
    queued_at: Range[datetime]
    """
    Include jobs queued within the given time range.
    """
    started_at: Range[datetime]
    """
    Include jobs started within the given time range.
    """
    waiting_at: Range[datetime]
    """
    Include jobs that started waiting within the given time range.
    """
    completed_at: Range[datetime]
    """
    Include jobs completed within the given time range.
    """
    killed_at: Range[datetime]
    """
    Include jobs killed within the given time range.
    """
    failed_at: Range[datetime]
    """
    Include jobs failed within the given time range.
    """
    exported_at: Range[datetime]
    """
    Include jobs exported within the given time range.
    """



import os
from io import StringIO
from pathlib import Path

import pytest

from cryosparc import star


@pytest.fixture
def simple_star_file():
    return StringIO(
        """
data_

loop_
_rlnCoordinateX #1
_rlnCoordinateY #2
3410.0 5184.0
680.0 6525.0
816.0 7370.0
515.0 7118.0
1088.0 7254.0
3138.0 7195.0
3381.0 7040.0
3507.0 7127.0
3867.0 7147.0
4012.0 7020.0
4178.0 7079.0
4450.0 6952.0
4119.0 6729.0
3711.0 6321.0
3245.0 6127.0
4314.0 6088.0
4877.0 5738.0
4974.0 5631.0
4362.0 4504.0
4207.0 4495.0
4051.0 4495.0
4168.0 4864.0
3585.0 4320.0
3216.0 4281.0
3225.0 4125.0
3012.0 3533.0
3274.0 3591.0
2011.0 3377.0
1632.0 3377.0
1632.0 3484.0

"""
    )


@pytest.fixture
def big_star_file():
    return Path() / "tests" / "data" / "job096_run_data.star"


def test_read_simple(simple_star_file):
    data = star.read(simple_star_file)
    assert "" in data
    assert len(data[""]) == 30
    assert data[""]["rlnCoordinateX"][0] == 3410.0
    assert data[""]["rlnCoordinateY"][0] == 5184.0
    assert data[""]["rlnCoordinateX"][-1] == 1632.0
    assert data[""]["rlnCoordinateY"][-1] == 3484.0


@pytest.mark.skipif(os.getenv("CI") == "true", reason="To avoid GitHub LFS download on GitHub Actions")
def test_read_big(big_star_file):
    data = star.read(big_star_file)
    assert "optics" in data
    assert "particles" in data
    assert len(data["optics"]) == 1
    assert len(data["optics"][0]) == 13
    assert len(data["particles"]) == 29183
    assert len(data["particles"][0]) == 26
    assert data["optics"]["rlnOpticsGroupName"][0] == "mydata"
    assert data["optics"]["rlnOpticsGroup"][0] == 1
    assert data["particles"]["rlnCoordinateX"][0] == 2688.0
    assert data["particles"]["rlnCoordinateY"][0] == 1330.0
    assert data["particles"]["rlnImageName"][0] == "1@Polished/SARS2_HGKA54_ZK144_G2_K3_0000_shiny.mrcs"
    assert data["particles"]["rlnRandomSubset"][0] == 1
    assert data["particles"]["rlnCoordinateX"][-1] == 4221.0
    assert data["particles"]["rlnCoordinateY"][-1] == 3374.0
    assert data["particles"]["rlnImageName"][-1] == "12@Polished/SARS2_HGKA54_ZK144_G2_K3_7982_shiny.mrcs"
    assert data["particles"]["rlnRandomSubset"][-1] == 2


def test_write():
    result = StringIO()
    star.write(
        result,
        data=[
            (123.0, 456.0, "particles.mrc", "mic.mrc", 1),
            (789.0, 123.0, "particles.mrc", "mic.mrc", 0),
        ],
        labels=[
            "rlnCoordinateX",
            "rlnCoordinateY",
            "rlnImageName",
            "rlnMicrographName",
            "rlnCtfDataAreCtfPremultiplied",
        ],
    )
    assert (
        result.getvalue()
        == """
data_

loop_
_rlnCoordinateX #1
_rlnCoordinateY #2
_rlnImageName #3
_rlnMicrographName #4
_rlnCtfDataAreCtfPremultiplied #5
123.0 456.0 particles.mrc mic.mrc 1
789.0 123.0 particles.mrc mic.mrc 0

"""
    )


def test_write_simple(simple_star_file):
    data = star.read(simple_star_file)
    result = StringIO()
    star.write_blocks(result, data)
    assert result.getvalue() == simple_star_file.getvalue()
